'''Specifies  a distributed AI setting.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..neighbor_manager.node_id import NodeID
from ..wrangler.constants import Defaults

from .impl.config_component import (
    ConfigComponent, ConfigError, InvalidValueError, ValidationErrors)
from .impl.config_component_catalog import ConfigComponentCatalog


class SplitConfig(ConfigComponent):
    '''Specifies how to split data for simulations.'''

    @property
    def seed(self) -> int:
        '''Seed for random number generation.'''
        return self._get_with_default('seed', dflt=Defaults.SEED)

    @property
    def num_nodes(self) -> int:
        '''The number of nodes in the system.'''
        return self._get('num_nodes')


class DeciderConfig(ConfigComponent):
    '''Specifies configuration for a single decider.'''
    _name = 'decider'
    tags: Dict[str, Any] = {}

    def __init__(self, name: str, clause: Dict[str, Any]) -> None:
        self._name = name
        super().__init__(clause)

    class Keys(AEnum):
        '''Valid keys for the top level.'''
        ENABLED = 'enabled'

    def required_keys(self) -> Set[str]:
        return set()

    @property
    def name(self) -> str:
        '''Name of the decider.'''
        return self._name

    @property
    def enabled(self) -> bool:
        '''Is the decider enabled?'''
        return self._get_with_default('enabled', dflt=True)


class DecidersConfig(ConfigComponent):
    '''Specifies configuration for all deciders.'''

    def __init__(self, clause: Dict[str, Any]) -> None:
        super().__init__(clause)

    @property
    def decider_configs(self) -> Dict[str, DeciderConfig]:
        '''Get all decider configurations.'''
        return {name: getattr(self, name) for name in self._clause.keys()}

    def required_keys(self) -> Set[str]:
        return set()


class CommunicatorConfig(ConfigComponent):
    '''Specifies configuration for a communicator.'''
    _my_id: NodeID

    def __init__(self, clause: Dict[str, Any], my_id: NodeID, **kwargs) -> None:
        super().__init__(clause=clause, my_id=my_id, **kwargs)
        self._my_id = my_id

    class Constants(Enum):
        '''Keys below the top level.'''
        NAME = 'name'
        STRATEGY = 'strategy'

    class Strategies(Enum):
        '''Strategies for the communicator.'''
        BROADCAST = 'broadcast'  # All messages go to all neighbors.
        UNICAST = 'unicast'  # Messages only go to qualifying neighbors.

    class Defaults(Enum):
        '''Default values'''
        COMMUNICATOR = 'memory'
        STRATEGY = 'broadcast'

    @property
    def name(self) -> str:
        '''Name of the communicator.

        Options include 'memory', 'sockets', and 'kafka'.
        '''
        return self._get_with_default(self.Constants.NAME, dflt=self.Defaults.COMMUNICATOR.value)

    @property
    def strategy(self) -> str:
        '''Strategy for the communicator.'''
        return self._get_with_default(self.Constants.STRATEGY, dflt=self.Defaults.STRATEGY.value)


class DropperConfig(ConfigComponent):
    '''Specifies configuration for packet dropper.'''
    name = 'dropper'
    tags: Dict[str, Any] = {}

    class Constants(Enum):
        '''Keys below the top level.'''

    class Keys(AEnum):
        '''Valid keys for the top level.'''
        DROP_RATE = 'drop_rate'
        OUTPUT_DIR = 'output_dir'
        SEED = 'seed'

    def required_keys(self) -> Set[str]:
        return {
            self.Keys.DROP_RATE.value,  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        }

    @property
    def drop_rate(self) -> float:
        '''Probability of dropping a packet.'''
        return float(self._get_with_default(self.Keys.DROP_RATE, dflt=0.0))

    @property
    def seed(self) -> int:
        '''Seed for random number generation.'''
        return self._get_with_default('seed', dflt=Defaults.SEED)

    @property
    def output_dir(self) -> Optional[Path]:
        '''The number of nodes in the system.'''
        retval = self._get_with_default(self.Keys.OUTPUT_DIR, dflt=None)
        return Path(retval) if retval is not None else None


class DistributedConfig(ConfigComponent):
    '''Specifies distributed configration in a distributed AI setting.'''
    name = 'distributed'
    tags: Dict[str, Any] = {}

    class Constants(Enum):
        '''Keys below the top level.'''
        NAME = 'name'
        # Discoverers
        STATIC = 'static'  # Static neighbor configuration
        ADJACENCY = 'adjacency'
        EDGES = 'edges'
        DYNAMIC = 'dynamic'  # Simple dynamic neighbor configuration
        # Subitems
        NODES_AND_ENDPOINTS = 'nodes_and_endpoints'
        # Regularization
        WEIGHT_MATRIX = 'weight_matrix'

    class Defaults(Enum):
        '''Default values'''
        DISCOVERER = 'dynamic'
        POLLING_INTERVAL = '0.1'
        FIT_EPS = '0.0000001'
        REGULARIZATION = 'function_space'
        PWFF_TIMEOUT = '60.000000000'  # seconds
        ONLY_SEND_PERIODICALLY = False
        CONVERGENCE_TIME_NS = 1000000000  # 1 second

    class Keys(AEnum):
        '''Valid keys for the top level.'''
        COMMUNICATOR = 'communicator'  # Config for the communicator to use
        DISCOVERER = 'discoverer'  # Config for the discoverer to use
        DECIDER = 'decider'  # Config for the decider to use
        DROPPER = 'dropper'  # Configuration for packet dropper
        FIT_EPS = 'fit_eps'  # Minimum change in model before sending to neighbors
        MY_ID = 'my_id'  # ID of the current node
        PIPELINES = 'pipelines'  # Clause of pipeline loaders
        POLLING_INTERVAL = 'polling_interval'  # Time for models to wait between refitting
        PWFF_TIMEOUT = 'predict_wait_for_fit_timeout'  # TO predict() to wait for fit() to complete
        REGULARIZATION = 'regularization'  # Info about nieghbor regularization
        SPLIT = 'split'  # Info about splitting of the data for simulations
        ONLY_SEND_PERIODICALLY = 'only_send_periodically'
        CONVERGENCE_TIME_NS = 'convergence_time_ns'

    def required_keys(self) -> Set[str]:
        return {
            self.Keys.DISCOVERER.value,  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            self.Keys.COMMUNICATOR.value  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        }

    def get_static_adjacency(self, my_id: NodeID) -> Optional[List[NodeID]]:
        '''If we have a static config, get the list of my_id's neighbors.

        Returns None if there is no static config.
        '''
        if not self._exists(self.Keys.DISCOVERER, self.Constants.STATIC):
            return None
        if self._get(self.Keys.DISCOVERER, self.Constants.NAME) != self.Constants.STATIC.value:
            return None
        if self._exists(self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.ADJACENCY):
            my_neighbors: List[int] = self._get(
                self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.ADJACENCY, str(my_id))
            return [NodeID(n) for n in sorted(my_neighbors)]
        if self._exists(self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.EDGES):
            retval = set()
            for edge in self._get(
                self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.EDGES
            ):
                if edge[0] == int(my_id):
                    retval.add(edge[1])
                if edge[1] == int(my_id):
                    retval.add(edge[0])
            return [NodeID(n) for n in sorted(retval)]
        raise NotImplementedError(
            f'unrecognized static clause(s): '
            f'{self._get(self.Keys.DISCOVERER, self.Constants.STATIC).keys()}')

    def get_adjacency(self) -> Optional[Dict[NodeID, List[NodeID]]]:
        '''If we have a static config, get full adjacency of the network.

        Returns None if there is no static config.
        '''
        if not self._exists(self.Keys.DISCOVERER, self.Constants.STATIC):
            return None
        if self._get(self.Keys.DISCOVERER, self.Constants.NAME) != self.Constants.STATIC.value:
            return None
        if self._exists(self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.ADJACENCY):
            adjacency = self._get(
                self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.ADJACENCY)
            return {NodeID(int(k)): [NodeID(n) for n in v] for k, v in adjacency.items()}
        if self._exists(self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.EDGES):
            edges = self._get(
                self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.EDGES)
            retval: Dict[NodeID, List[NodeID]] = {}
            for edge in edges:
                if edge[0] not in retval:
                    retval[edge[0]] = []
                if edge[1] not in retval:
                    retval[edge[1]] = []
                retval[edge[0]].append(edge[1])
                retval[edge[1]].append(edge[0])
            return {NodeID(int(k)): [NodeID(n) for n in v] for k, v in retval.items()}
        raise NotImplementedError(
            f'unrecognized static clause(s): '
            f'{self._get(self.Keys.DISCOVERER, self.Constants.STATIC).keys()}')

    def validate(self, **kwargs: Any) -> None:
        errors: List[ConfigError] = []

        if self._exists(self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.EDGES):
            edges = self._get(
                self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.EDGES)
            if not isinstance(edges, list):
                raise ValidationErrors([InvalidValueError(
                    f'edges must be a list, instead found {edges} of type {type(edges)}')])
            for edge in self._get(
                self.Keys.DISCOVERER, self.Constants.STATIC, self.Constants.EDGES
            ):
                if len(edge) != 2:
                    errors.append(InvalidValueError(
                        f'Edge {edge} does not have exactly two elements.'))

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)

    @property
    def fit_eps(self) -> float:
        '''Minimum change in model before sending to neighbors.'''
        return float(self._get_with_default(
            self.Keys.FIT_EPS, dflt=self.Defaults.FIT_EPS.value))

    @property
    def polling_interval(self) -> float:
        '''Time for models to wait between refitting in seconds, float.'''
        return float(self._get_with_default(
            self.Keys.POLLING_INTERVAL, dflt=self.Defaults.POLLING_INTERVAL.value))

    @property
    def decider(self) -> DecidersConfig:
        '''Configuration for the deciders.'''
        return DecidersConfig(clause=self._get_with_default(
            self.Keys.DECIDER, dflt={}))

    @property
    def no_deciders(self) -> bool:
        '''Are there no enabled deciders?'''
        for decider in self.decider.decider_configs.values():
            if decider.enabled:
                return False
        return True

    @property
    def discoverer(self) -> str:
        '''Which discoverer should we use?

        Options include 'static' and 'dynamic'.
        '''
        return self._get_with_default(
            self.Keys.DISCOVERER, self.Constants.NAME, dflt=self.Defaults.DISCOVERER.value)

    @property
    def communicator(self) -> CommunicatorConfig:
        '''Configuration for the communicator.'''
        return CommunicatorConfig(clause=self._get_with_default(self.Keys.COMMUNICATOR, dflt={}),
                                  my_id=self.my_id)

    @property
    def dropper(self) -> Optional[DropperConfig]:
        '''Configuration for packet dropper.'''
        clause = self._get_with_default(self.Keys.DROPPER, dflt=None)
        return DropperConfig(clause) if clause is not None else None

    @property
    def split(self) -> Optional[SplitConfig]:
        '''Info about splitting of the data for simulations.'''
        clause = self._get_with_default(self.Keys.SPLIT, dflt=None)
        if clause is None:
            return None
        assert isinstance(clause, dict), (
            f'BUG: expected a dict at {self.Keys.SPLIT}, instead found {type(clause)}'
        )

        return SplitConfig(clause)

    @property
    def my_id(self) -> NodeID:
        '''Node ID for the current node.'''
        # TODO(Piggy): rewrite this, injecting my_id from the command line.
        return self._get_with_default(self.Keys.MY_ID, dflt=NodeID(1))

    @my_id.setter
    def my_id(self, value: NodeID) -> None:
        '''Set the node ID for the current node.'''
        self._clause[self.Keys.MY_ID.value] = value  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long

    @property
    def only_send_periodically(self) -> bool:
        '''If true, we do not consider sending on every fit.'''
        return self._get_with_default(
            self.Keys.ONLY_SEND_PERIODICALLY, dflt=self.Defaults.ONLY_SEND_PERIODICALLY.value)

    @property
    def pipelines(self) -> Dict[str, List[Any]]:
        '''List of pipeline loaders to use.'''
        return self._get_with_default(self.Keys.PIPELINES, dflt={})

    @property
    def regularization_type(self) -> str:
        '''Type of neighbor regularization'''
        return self._get_with_default(self.Keys.REGULARIZATION, self.Constants.NAME,
                                      dflt=self.Defaults.REGULARIZATION)

    @property
    def neighbor_weights(self) -> Dict[NodeID, float]:
        '''Weights for neighbors, if used in regularization.'''
        # TODO(merritt): relies on node IDs starting at 0
        matrix = self._get_with_default(self.Keys.REGULARIZATION, self.Constants.WEIGHT_MATRIX,
                                        dflt=None)
        if matrix is None:
            return {}

        row = matrix[int(self.my_id)]
        return {NodeID(i): weight for i, weight in enumerate(row)}

    @property
    def predict_wait_for_fit_timeout_ns(self) -> int:
        '''How long should predict() wait for fit() to complete?  (In nanoseconds)'''
        pwff_s = float(self._get_with_default(self.Keys.PWFF_TIMEOUT,
                                              dflt=self.Defaults.PWFF_TIMEOUT.value))
        return int(1000000000 * pwff_s)

    @property
    def convergence_time_ns(self) -> int:
        '''How long should we wait for convergence?  (In nanoseconds)'''
        return int(self._get_with_default(self.Keys.CONVERGENCE_TIME_NS,
                                          dflt=self.Defaults.CONVERGENCE_TIME_NS.value))


def register(catalog: ConfigComponentCatalog) -> None:
    '''Register all the objects in this file.'''
    catalog.register(DistributedConfig)



'''tests for splitter.py'''
# pylint: disable=missing-function-docstring,missing-class-docstring
# pylint: disable=protected-access,duplicate-code

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...problem_def.cross_validation_config import CrossValidationConfig
from ...wrangler.dataset import Dataset
from .splitter import Splitter, SplitDataset


class FakeSplitter(Splitter):
    _name = 'fake'
    _hyperparams = {
        'k1': 'v1',
        'k2': 'v2'
    }

    def split(self,
              dataset: Dataset,
              **unused_kwargs) -> SplitDataset:
        return SplitDataset()


def test_hyperparam_overrides() -> None:
    dut = FakeSplitter(cv_config=CrossValidationConfig({}))
    assert dut._hyperparams == {
        'k1': 'v1',
        'k2': 'v2'
    }
    assert dut.hyperparams(k2='new_val', new_key='new_val2') == {
        'k1': 'v1',
        'k2': 'new_val',
        'new_key': 'new_val2'
    }



'''Tests from ColumnParser'''
# pylint: disable=missing-function-docstring, duplicate-code
import os
from pathlib import Path
from typing import Dict, Any

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd

from .column_parser import ColumnParser
from ..wrangler.dataset import Dataset


def valid_csv() -> str:
    '''Returns a path (in the form of a string) to a valid csv file.'''
    current_path = str(os.getenv('PYTEST_CURRENT_TEST')).split('::', maxsplit=1)[0]
    pathobj = Path(current_path).resolve()
    module_parent = pathobj.parents[2]
    path = module_parent / 'examples' / 'regression' / 'diabetes.csv'
    return str(path)


def make_dataset_clause() -> Dict[str, Any]:
    retval = {
        'config': 'local',
        'train_path': valid_csv(),
        'column_roles': {
            'target': {
                'name': 'Prog'
            },
            'index': {
                'id': 0
            }
        }
    }
    return retval


def test_parse_trivial_dataset() -> None:
    triv = Dataset()
    triv.dataframe = pd.DataFrame({
        'a': ['cat_one', 'cat_two', 'cat_two'],
        'b': [22.1, 22.3, 22.7]})
    assert triv.dataframe.dtypes['a'] == 'object'
    dut = ColumnParser().instantiate()
    got = dut.predict(dataset=triv)
    assert got.dataframe.dtypes['a'] == 'category'



'''Test the metric catalog'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .metric import MetricStub
from .metric_catalog import MetricCatalogStub

# pylint: disable=missing-function-docstring,duplicate-code


def test_register() -> None:
    dut = MetricCatalogStub()
    dummy_metric = MetricStub()
    assert 'stub_metric' == dut.register(obj=dummy_metric)


def test_register_name() -> None:
    dut = MetricCatalogStub()
    dummy_metric = MetricStub()
    assert 'somenamedmetric' == dut.register(obj=dummy_metric, name='SomeNamedMetric')


def test_lookup_by_name() -> None:
    dut = MetricCatalogStub()
    some_metric = MetricStub()
    dut.register(obj=some_metric, name='SomeNamedMetric')
    assert dut.lookup_by_name('SomeNamedMetric') == some_metric


def test_lookup_one_by_tag() -> None:
    dut = MetricCatalogStub()
    dummy_metric = MetricStub()
    dut.register(obj=dummy_metric, name='NamedDummyMetric', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 1


def test_lookup_by_tag_fail() -> None:
    dut = MetricCatalogStub()
    dummy_metric = MetricStub()
    dut.register(obj=dummy_metric, name='NamedDummyMetric', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='different_value')) == 0


def test_lookup_multi_by_tag() -> None:
    dut = MetricCatalogStub()
    dummy_metric = MetricStub()
    dut.register(obj=dummy_metric, name="Name1", tags={'k1': ['v1']})
    dummy_metric2 = MetricStub()
    dut.register(obj=dummy_metric2, name="Name2", tags={'k1': ['v1']})
    dummy_metric3 = MetricStub()
    dut.register(obj=dummy_metric3, name="Name3", tags={'k1': ['different_value']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 2



'''Object that listens for inbound messages and sends outbound messages'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.
import abc
from typing import Dict, List, Mapping, Optional

from ...algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ...config_components.distributed_config import DeciderConfig
from ...neighbor_manager.node_id import NodeID


class Decider(metaclass=abc.ABCMeta):
    '''Object that decides if we should send our state to a neighbor.'''
    name: str
    tags: Dict[str, List[str]]
    _config: Optional[DeciderConfig] = None

    # All subclasses must implement this __init__ signature.
    # When we get something from a catalog we don't otherwise know what
    # arguments it needs.
    def __init__(self,
                 name: Optional[str] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 config: Optional[DeciderConfig] = None
                 ) -> None:
        self._config = config
        self.name = name if name is not None else self.name
        self.tags = tags if tags is not None else self.tags

    @abc.abstractmethod
    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''



'''File to  explore https://gitlab.com/autonlab/ngautonml/-/issues/1'''

import code
import logging
import warnings

from ..problem_def.problem_def import ProblemDefinition
from ..wrangler.wrangler import Wrangler

logging.basicConfig(level=logging.ERROR)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

if __name__ == '__main__':
    pdef = ProblemDefinition('''
    {
        "_comments" : [
            "A json file fully encapsulating the problem definition for openml dataset #31.",
            "This dataset is a tabular binary classification problem.",
            "People are classified as good or bad credit risks based on attributes."
        ],
        "dataset" : {
            "config" : "local",
            "train_path" : "examples/classification/credit-train.csv",
            "test_path": "examples/classification/credit-test.csv",
            "column_roles": {
                "target" : {
                    "name" : "class",
                    "pos_label": "good"
                }
            }
        },
        "problem_type" : {
            "data_type": "TABULAR",
            "task": "BINARY_CLASSIFICATION"
        },
        "metrics" :  {
            "roc_auc_score": {},
            "accuracy_score": {}
        },
        "output" : {}
    }
    ''')

    dut = Wrangler(problem_definition=pdef)

    got = dut.fit_predict_rank()

    code.interact(local=locals())



'''Holds an instance of a model.'''
import abc
from typing import Any, Dict, Optional
from typing_extensions import Protocol

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ...wrangler.dataset import Dataset

from .algorithm import Algorithm


class Error(Exception):
    '''Base error class for AlgorithmInstance.'''


class DatasetError(Error):
    '''Something is malformed about the dataset.'''


class AlgorithmInstance(metaclass=abc.ABCMeta):
    '''Holds an instance of an algorithm as made by the instantiator.'''
    _algorithm: Optional[Algorithm]
    _hyperparameters: Dict[str, Any]

    def __init__(self, parent: Algorithm, **kwargs):
        self._algorithm = parent
        self._hyperparameters = kwargs

    @property
    def algorithm(self) -> Algorithm:
        '''self._algorithm asserted to non-optional'''
        assert self._algorithm is not None, (
            'BUG: we should never access AlgorithmInstance.algorithm if it is not set.'
        )
        return self._algorithm

    def hyperparams(self, **kwargs) -> Dict[str, Any]:
        '''Return hyperparameters for this instance.'''
        retval = self._hyperparameters.copy()
        retval.update(kwargs)
        if self._algorithm is None:
            return retval
        return self._algorithm.hyperparams(**retval)

    @abc.abstractmethod
    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        '''Apply model to input dataset to create output.

        This may require that the model is fit (self.trained == True) before it is called.
        '''

    @property
    def catalog_name(self) -> str:
        '''The catalog name of our algorithm.'''
        assert self._algorithm is not None
        return self._algorithm.name

    def start(self) -> None:
        '''Start any threads that the algorithm needs.

        This is a no-op if the algorithm does not have threads.
        '''

    def stop(self) -> None:
        '''Stop all the threads.

        This is a no-op if the algorithm does not have threads.
        '''

    @property
    def locked(self) -> bool:
        '''Is this instance currently thread-locked?

        This can only return True for subclasses that spawn their own threads.
        '''
        return False

    @property
    def converged(self) -> bool:
        '''If this is a distributed model, has it converged?

        Always returns True for non-distributed models.
        '''
        return True


class Constructor(Protocol):
    '''Match the signature of a generic model constructor.'''
    def __call__(self, **kwargs: Any) -> Any:
        ...


def not_implemented(**kwargs):
    '''A constructor must be specified in a subclass'''
    raise NotImplementedError('Implementation must specify a model _constructor')



'''Tests for Algorithm object'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional
import pytest

from ...problem_def.task import TaskType
from ...wrangler.dataset import Dataset
from .algorithm import Algorithm, AlgorithmCatalogStub
from .algorithm_instance import AlgorithmInstance

# pylint: disable=missing-function-docstring,missing-class-docstring
# pylint: disable=protected-access, duplicate-code


class FakeAlgorithmInstance(AlgorithmInstance):
    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if dataset is None:
            return None
        return Dataset(dataset.copy())


class FakeAlgorithm(Algorithm):
    _name = 'fake algorithm'
    _default_hyperparams = {
        'base_param': 'base_value',
        'param_to_override': 'value_to_override',
    }
    _tags = {
        'task': [
            TaskType.REGRESSION.name,
            TaskType.BINARY_CLASSIFICATION.name,
        ]
    }
    _hyperparam_lookup = {
        'typed_param': {
            '<class \'float\'>': float,
            '<class \'int\'>': int,
        }
    }

    def instantiate(self, **hyperparams) -> FakeAlgorithmInstance:
        return FakeAlgorithmInstance(parent=self, **hyperparams)


class FakeAlgorithm2(Algorithm):
    _name = 'fake algorithm with different tags'
    _default_hyperparams = {
        'base_param': 'base_value',
        'param_to_override': 'value_to_override',
    }
    _tags = {
        'task': [
            TaskType.BINARY_CLASSIFICATION.name,
            TaskType.MULTICLASS_CLASSIFICATION.name,
        ]
    }

    def instantiate(self, **hyperparams) -> FakeAlgorithmInstance:
        return FakeAlgorithmInstance(parent=self, **hyperparams)


def test_hyperparams_at_construction() -> None:
    algorithm = FakeAlgorithm(param_to_override='overridden_value')
    assert algorithm.hyperparams(instantiation_time_param='instantiation_value') == {
        'base_param': 'base_value',
        'param_to_override': 'overridden_value',
        'instantiation_time_param': 'instantiation_value',
    }

    algorithm2 = FakeAlgorithm()
    assert algorithm2.hyperparams() == {
        'base_param': 'base_value',
        'param_to_override': 'value_to_override',
    }


def test_algorithm_catalog_basic() -> None:
    dut = AlgorithmCatalogStub()
    fm1 = FakeAlgorithm()
    fm2 = FakeAlgorithm()
    output = dut.register(obj=fm1)
    print(f'register output: {output}')
    assert output == 'fake algorithm'
    assert dut.register(obj=fm2, name='override name') == 'override name'
    assert dut.lookup_by_name('fake algorithm') == fm1
    assert dut.lookup_by_name('override name') == fm2


def test_algorithm_catalog_lookup_by_tags() -> None:
    dut = AlgorithmCatalogStub()
    fm1 = FakeAlgorithm()
    fm2 = FakeAlgorithm2()
    dut.register(fm1)
    dut.register(fm2)
    assert len(dut.lookup_by_tag_and(
        task=TaskType.BINARY_CLASSIFICATION.name)) == 2
    assert len(dut.lookup_by_tag_and(
        task=TaskType.MULTICLASS_CLASSIFICATION.name)) == 1
    assert len(dut.lookup_by_tag_and(
        task=TaskType.FORECASTING.name)) == 0


def test_algorithm_catalog_get_tag_types() -> None:
    dut = AlgorithmCatalogStub()
    dut.register(FakeAlgorithm())
    assert dut.tagtypes == {'task'}


def test_algorithm_catalog_get_tag_values() -> None:
    dut = AlgorithmCatalogStub()
    dut.register(FakeAlgorithm())
    dut.register(FakeAlgorithm2())
    want = {TaskType.BINARY_CLASSIFICATION.name.lower(),
            TaskType.MULTICLASS_CLASSIFICATION.name.lower(),
            TaskType.REGRESSION.name.lower()}
    assert dut.tagvals('task') == want


def test_algorithm_parse_and_repr() -> None:
    dut = FakeAlgorithm()
    assert dut._param_from_json('typed_param', '<class \'float\'>') == float
    assert dut.param_to_json('typed_param', float) == '<class \'float\'>'


def test_algorithm_parse_and_repr_default() -> None:
    dut = FakeAlgorithm()
    assert dut._param_from_json('some_param', 'some_value') == 'some_value'
    assert dut.param_to_json('some_param', 'some_value') == 'some_value'


def test_algorithm_parse_fail() -> None:
    dut = FakeAlgorithm()
    with pytest.raises(KeyError, match='typed_param'):
        _ = dut._param_from_json('typed_param', 'missing_value')


def test_algorithm_repr_fail() -> None:
    dut = FakeAlgorithm()
    with pytest.raises(KeyError, match='typed_param'):
        _ = dut.param_to_json('typed_param', 'missing_value')



'''A NeighborState subclass for testing purposes.'''
# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import List, Optional

from .distributed_algorithm_instance import NeighborState


class ToyNeighborState(NeighborState):
    '''Test class for NeighborState.'''
    _center: float

    def __init__(self, columns: Optional[List[str]] = None, center: float = 0.0) -> None:
        super().__init__(columns=columns)
        self._center = center

    @property
    def center(self) -> float:
        '''Get the center.'''
        return self._center

    def encode(self) -> bytes:
        '''Encode a message for distributed neighbors. '''
        return b'test neighbor state'

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'ToyNeighborState':
        '''Decode a message from distributed neighbors.

        This should just redirect to the relevant NeighborState subclass.
        '''
        return cls(center=0.0)

    def distance(self, other: NeighborState) -> float:
        '''Return the distance between two states.'''
        assert isinstance(other, ToyNeighborState)
        return abs(self.center - other.center)



'''Generic NeighborManager for any communicator and discoverer'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


import logging
import queue
from typing import List, Optional

from ..communicators.impl.communicator import Communicator
from ..discoverers.impl.discoverer import Discoverer
from ..wrangler.logger import Logger
from .neighbor_manager import Event, NeighborManager
from .node_id import NodeID

log = Logger(__file__, level=logging.DEBUG).logger()


class NeighborManagerImpl(NeighborManager):
    '''Generic neighbor manager with configurable discovery.'''
    name = 'neighbor_manager_impl'
    tags = {}
    _discoverer: Discoverer
    _communicator: Communicator

    def __init__(self, discoverer: Discoverer, **kwargs):
        super().__init__(discoverer=discoverer, **kwargs)
        self._discoverer = discoverer
        # Set up a socket to listen to.
        self._communicator = self._discoverer.communicator

    def synchronize(self):
        # TODO(Merritt/Piggy): write this
        return

    def _poll_for_events(self, timeout: Optional[float] = None) -> List[Event]:
        retval = []
        # We do not use self._default_timeout for polling. If you want
        # poll_for_events to block, you must specify a timeout.
        try:
            while not self._queue.empty():
                # There is a race here between callers to poll_for_events().
                # One of the racers will raise a queue.Empty, which is safely
                # caught below.
                event = self._queue.get(block=False)
                retval.append(event)
            if timeout is not None:
                # If we have a timeout, wait the full timeout for another event.
                event = self._queue.get(timeout=timeout)
                retval.append(event)
        except queue.Empty:
            pass
        log.debug('poll_for_events got events: %s', retval)
        return retval

    def _send_all(self, payload: bytes) -> None:
        log.debug('_send_all(%s)', payload[:50])
        self._communicator.send_all(payload)

    def _send(self, nodes: List[NodeID], payload: bytes) -> None:
        log.debug('_send: %s to %s', payload[:50], nodes)
        self._communicator.send(nodes, payload)



'''Decider that unconditionally returns True.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from copy import deepcopy
from typing import Mapping

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig, DeciderConfig
from ..neighbor_manager.node_id import NodeID

from .impl.decider import Decider
from .impl.decider_catalog import DeciderCatalog


class TrueDeciderConfig(DeciderConfig):
    '''Configuration for TrueDecider.'''

    def __init__(self, name: str, clause: dict) -> None:
        subclause = deepcopy(clause)
        subclause.pop(TrueDeciderConfig.Keys.ENABLED, None)
        if subclause:
            raise ValueError('TrueDeciderConfig takes only an enabled key.')
        super().__init__(name=name, clause=clause)

    def __str__(self) -> str:
        return f'true({",".join([str(c) for c in self._clause])})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TrueDeciderConfig):
            return False
        return True

    class Keys(AEnum):
        '''Keys for the clause.'''
        ENABLED = 'enabled'


class TrueDecider(Decider):
    '''Decider that decides True unconditionally.'''
    name = 'true'
    tags = {}

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        return True


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(TrueDecider)
    DecidersConfig.register(name=TrueDecider.name, config_type=TrueDeciderConfig)



'''Catalog for objects that control neighbor discovery in distributed algorithms.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Dict, List, Optional, Type, Union

from ...catalog.memory_catalog import MemoryCatalog

from .discoverer import Discoverer


class DiscovererCatalog(MemoryCatalog[Type[Discoverer]], metaclass=abc.ABCMeta):
    '''Base class for Discoverer catalogs'''

    def register(self, obj: Type[Discoverer], name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        return super().register(obj, name, tags)


class DiscovererCatalogStub(DiscovererCatalog):
    '''stub'''



'''Loads a local file as a pandas DataFrame.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from scipy.io.arff import loadarff  # type: ignore[import]

from ..config_components.dataset_config import DatasetConfig
from ..wrangler.constants import FileType
from ..wrangler.dataset import Dataset, DatasetKeys, RoleName, TaskType
from ..wrangler.logger import Level, Logger
from .impl.dataframe_loader import DataframeLoader, DatasetFileError
from .impl.data_loader_catalog import DataLoaderCatalog

logger = Logger(__file__, level=Level.DEBUG).logger()


class LocalDataLoader(DataframeLoader):
    '''Loads a local file as a pandas DataFrame.'''

    name: str = 'local'
    tags: Dict[str, List[str]] = {
        'input_format': ['csv', 'arff', 'tabular_file'],
        'loaded_format': ['pandas_dataframe'],
    }

    _train_path: Path
    _test_path: Optional[Path] = None
    _ext: str

    def __init__(self, config: DatasetConfig):
        train_path: Path = Path(config.params[config.Keys.TRAIN_PATH.value])  # type: ignore[attr-defined] # pylint: disable=line-too-long
        self._train_path = train_path.expanduser().resolve()
        ext: str = train_path.suffix[1:]

        if ext not in FileType.list():
            raise DatasetFileError(
                f'Train path from the problem definition ({train_path})'
                f' does not point to a file with one of these extensions: {FileType.list()}'
            )

        self._ext = ext

        if config.Keys.TEST_PATH.value in config.params:  # type: ignore[attr-defined]
            test_path = Path(config.params[config.Keys.TEST_PATH.value])  # type: ignore[attr-defined] # pylint: disable=line-too-long
            test_path = test_path.expanduser().resolve()
            test_ext = test_path.suffix[1:]
            if test_ext != ext:
                raise DatasetFileError(
                    f'test data: "{test_path}" is not same file type as train data'
                    f'(Expected: .{ext})')

            self._test_path = test_path

        self._static_path = None
        if config.Keys.STATIC_EXOGENOUS_PATH.value in config.params:  # type: ignore[attr-defined]
            static_path = Path(
                config.params[config.Keys.STATIC_EXOGENOUS_PATH.value]).expanduser().resolve()  # type: ignore[attr-defined] # pylint: disable=line-too-long
            static_ext = static_path.suffix[1:]
            if static_ext != ext:
                raise DatasetFileError(
                    f'static data: "{static_path}" is not same file type as train data'
                    f'(Expected: .{ext})')

            self._static_path = static_path

        super().__init__(config=config)

    def _drop_unnamed_cols(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        '''Deal with pandas CSV loader loading the index as an unnamed column.

        There are a few potential ways to do this.
        We choose to drop all cols called 'Unnamed: [int of any length]',
        as those were likely unnamed in the actual csv.'''
        # TODO(Merritt): warn user if we encounter this under certain circumstances
        # (for ex: not in first column
        # or the col is literally named 'Unnamed: [num]' in the csv)
        drop_cols = dataframe.filter(regex=r'^Unnamed: \d+$').columns
        retval = dataframe.drop(labels=drop_cols, axis=1)
        return retval

    def _load_dataframe(self, path: Path) -> pd.DataFrame:
        '''Load the file at path based on self._ext'''
        retval = None

        try:
            if self._ext == FileType.CSV.value:
                retval = pd.read_csv(path)
            elif self._ext == FileType.ARFF.value:
                raw_data = loadarff(path)
                retval = pd.DataFrame(raw_data[0])
        except FileNotFoundError as exc:
            raise DatasetFileError(f'Unable to load file from "{path}".') from exc

        if retval is None:
            raise DatasetFileError(f'Unsupported file type: {path}')

        return self._drop_unnamed_cols(retval)

    def _load_dataset(self, path: Path) -> Dataset:
        '''Load a Dataset into a DataFrame from the provided Path'''
        retval = Dataset(metadata=self._metadata)

        data = self._load_dataframe(path=path)

        retval.dataframe = data

        # if this is a forecasting problem, load static exogenous data or create an empty key for it
        # TODO(Piggy/Merritt): replace this with something that can be used by both memory and local
        # to load additional data requested by a plugin
        if self._metadata.task == TaskType.FORECASTING:
            if self._static_path is None:
                retval[DatasetKeys.STATIC_EXOGENOUS.value] = None
                logger.debug('Forecasting problem has no static exogenous table.')
            else:
                static = self._load_dataframe(self._static_path)
                retval[DatasetKeys.STATIC_EXOGENOUS.value] = static

        return retval

    def _load_train(self) -> Dataset:
        return self._load_dataset(
            path=self._train_path)

    def _load_test(self) -> Optional[Dataset]:
        if self._test_path is None:
            return None

        retval = self._load_dataset(
            path=self._test_path)

        return retval

    def dataset(self,
                data: Any,
                key: Union[DatasetKeys, str] = DatasetKeys.DATAFRAME,
                cols: Optional[List[str]] = None,
                roles: Optional[List[Union[RoleName, str]]] = None,
                **kwargs) -> Dataset:
        '''Load a Dataset object, by placing data at the supplied key.

        As of 2023-12-04 we only know how to handle things we can turn into a pandas DataFrame.

        Args:
          :data: Dataframe or object that can be turned into one.
            If data is a Path or a str, will attempt
            to load the file it points to as a dataframe and put that in the dataset.
          :key: The key for the data in the dataset. Defaults to "dataframe".
          :cols: If set, only selects supplied column(s).
            Will take union with roles if both are specified.
          :roles: If set, only selects columns with supplied role(s).
            Will take union with cols if both are specified.
        '''

        if isinstance(data, (Path, str)):
            path = Path(data)
            data = self._load_dataframe(path=path)

        return super().dataset(data, key=key, cols=cols, roles=roles, **kwargs)


def register(catalog: DataLoaderCatalog):
    '''Register all the objects in this file.'''
    catalog.register(LocalDataLoader)



'''Wrappers of metrics from sklearn, currently only contains roc_auc_score.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import sklearn.metrics  # type: ignore[import]
from sklearn.metrics import mean_squared_error  # type: ignore[import]

from ..problem_def.task import Task, TaskType
from ..wrangler.dataset import Dataset, DatasetKeyError, DatasetValueError, RoleName
from .impl.metric import MetricInvalidDatasetError, SupervisedMetric
from .impl.metric_catalog import MetricCatalog


class SklearnMetric(SupervisedMetric):
    '''Wrapper for sklean.metrics'''
    _metric: Optional[Callable[..., float]] = None

    def __init__(self,
                 name: str,
                 metric: Optional[Callable[..., float]] = None,
                 tasks: Optional[List[TaskType]] = None,
                 high: bool = True,
                 needs_proba: bool = False,
                 needs_pos_label: bool = False,
                 implements_calculate: bool = True):
        if metric is None:
            metric = getattr(sklearn.metrics, name, None)
        self._metric = metric
        if tasks is None:
            tasks = []
        self._tags = {
            Task.Keys.TASK_TYPE.value: [t.name.lower() for t in tasks],  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        }
        self._tags['needs_pos_label'] = [str(needs_pos_label).lower()]
        self._tags['high'] = [str(high).lower()]
        self._tags['needs_proba'] = [str(needs_proba).lower()]
        self._tags['implements_calculate'] = [str(implements_calculate).lower()]
        super().__init__(name=name)

    def _calculate(self,
                   pred: Dataset,
                   ground_truth: Optional[Dataset] = None) -> Union[
                       float, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        # TODO(Merritt): should be able to accept hyperparams from problem def
        ground_truth = self.validate_input(pred, ground_truth)
        # validate_input is guaranteed to return ground truth Dataset rather than None.
        assert ground_truth.metadata.target is not None
        ground_truth_for_sklearn = list(
            ground_truth.ground_truth[ground_truth.metadata.target.name])
        assert pred.metadata.target is not None, (
            'BUG: calculate() called with no target info in metadata'
        )

        if self.needs_proba:
            try:
                if pred.probabilities.shape[1] == 2:
                    # Assume binary classification

                    # requires that binary classification target values are always [0, 1]
                    if set(ground_truth_for_sklearn) != {0, 1}:
                        raise MetricInvalidDatasetError(
                            f'{self._name} requires binary classification '
                            'target values to be [0, 1].  '
                            f'Instead found: {set(ground_truth_for_sklearn)}.')
                    # requires that probabilities gets output with the probabilities for 0 in the
                    # 0th column and 1 in the 1st column
                    #   and the probas for 1 in the 1th column
                    predictions_for_sklearn = pred.probabilities.as_(  # type: ignore[attr-defined]
                        np.ndarray)[:, 1]
                else:
                    predictions_for_sklearn = pred.probabilities.as_(  # type: ignore[attr-defined]
                        np.ndarray)
            except DatasetKeyError:
                # We are not using 'from' because this error will be displayed
                #   alongside metric scores as a explanation for why this score
                #   cannot be calculated, thus we want it to be succint.
                raise MetricInvalidDatasetError(   # pylint: disable=raise-missing-from
                    'No class probabilities output.')
            except DatasetValueError as err:
                raise MetricInvalidDatasetError(
                    'Expected DataFrame for probabilites.') from err
        else:
            pred_df = pred.predictions
            predictions_for_sklearn = np.array(
                pred_df[pred.metadata.target.name])

        assert self._metric is not None

        if RoleName.TARGET in pred.metadata.pos_labels and self.needs_pos_label:
            return self._metric(
                ground_truth_for_sklearn,
                predictions_for_sklearn,
                pos_label=pred.metadata.pos_labels[RoleName.TARGET])

        return self._metric(ground_truth_for_sklearn, predictions_for_sklearn)

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        got = self._calculate(pred, ground_truth)
        if not isinstance(got, float):
            raise NotImplementedError(
                f'Metric {self.name} does not return a float.  Instead found: {got}'
            )
        return got

    def calculate_roc_curve(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> Dataset:
        '''Calculate the ROC Curve.'''
        got = self._calculate(pred, ground_truth)
        if not isinstance(got, tuple) or len(got) != 3:
            raise NotImplementedError(
                f'Metric {self.name} does not return an roc curve of the expected format.  '
                f'Instead found: {got}'
            )
        tpr, fpr, thresholds = got
        retval = pred.output()
        retval.dataframe = pd.DataFrame({'fpr': fpr, 'tpr': tpr, 'thresholds': thresholds})
        return retval


def root_mean_squared_error(*args, **kwargs):
    '''Root Mean Squared Error'''
    return np.sqrt(mean_squared_error(*args, **kwargs))


METRICS: List[Dict[str, Any]] = [
    {
        'name': 'roc_auc_score',
        'tasks': [TaskType.BINARY_CLASSIFICATION],
        'needs_proba': True,
    },
    {
        'name': 'roc_curve',
        'tasks': [TaskType.BINARY_CLASSIFICATION],
        'needs_proba': True,
        'needs_pos_label': True,
        'implements_calculate': False
    },
    {
        'name': 'accuracy_score',
        'tasks': [TaskType.BINARY_CLASSIFICATION, TaskType.MULTICLASS_CLASSIFICATION],
    },
    {
        'name': 'f1_score',
        'high': True,
        'tasks': [TaskType.BINARY_CLASSIFICATION],
        'needs_pos_label': True,
    },
    {
        'name': 'mean_squared_error',
        'tasks': [TaskType.FORECASTING, TaskType.REGRESSION],
        'high': False,
    },
    {
        'name': 'root_mean_squared_error',
        'metric': root_mean_squared_error,
        'tasks': [TaskType.FORECASTING, TaskType.REGRESSION],
        'high': False,
    },
    {
        'name': 'mean_absolute_error',
        'tasks': [TaskType.FORECASTING, TaskType.REGRESSION],
        'high': False,
    },
    {
        'name': 'r2_score',
        'high': True,
        'tasks': [TaskType.REGRESSION],
    },
]


def register(catalog: MetricCatalog):
    '''Register all the metrics in this file.'''
    for met in METRICS:
        catalog.register(SklearnMetric(**met))



'''Tests for distributed_wrangler.py.'''

from copy import deepcopy
from multiprocessing import Process
import pickle
import queue
import select
import socket
import time
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import pytest
import requests
from sklearn import datasets  # type: ignore[import]

from ..algorithms.impl.synchronous import advance
from ..conftest import Clearer, Waiter
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import PipelineResults
from ..metrics.impl.metric_auto import MetricCatalogAuto
from ..problem_def.problem_def import ProblemDefinition
from ..tables.impl.table_auto import TableCatalogAuto
from ..wrangler.base_port import BasePort
from ..wrangler.dataset import df_to_dict_json_friendly
from ..wrangler.logger import Logger

from .distributed_wrangler import DistributedWrangler

# pylint: disable=missing-function-docstring, missing-class-docstring, duplicate-code, unused-variable
# pylint: disable=too-many-locals,no-member
# pylint: disable=duplicate-code, redefined-outer-name

log = Logger(__file__).logger()
_ = TableCatalogAuto()  # pylint: disable=pointless-statement
base_port = BasePort()


def load_classification_dataframes(
        reduced_train: bool = False) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    # Load the breast cancer dataset
    bc_x_full, bc_y_series = datasets.load_breast_cancer(return_X_y=True, as_frame=True)
    assert isinstance(bc_x_full, pd.DataFrame)
    assert isinstance(bc_y_series, pd.Series)
    bc_y = pd.DataFrame({'target': bc_y_series})

    # restrict number of attributes for wider variability of results
    bc_x = bc_x_full.iloc[:, :3]

    test_size = 50
    # Split the data into training/testing sets
    bc_x_train = bc_x[:-test_size]
    bc_x_test = bc_x[-test_size:].reset_index(drop=True)

    # Split the targets into training/testing sets
    bc_y_train = bc_y[:-test_size]
    bc_y_test = bc_y[-test_size:].reset_index(drop=True)

    train_df = pd.concat([bc_x_train, bc_y_train], axis=1)
    test_df = bc_x_test
    gt_df = bc_y_test

    if reduced_train:
        train_df = train_df.iloc[516:519].copy()
        train_df['target'] = [1, 0, 0]

    return (train_df, test_df, gt_df)


LOOPBACK = '127.0.0.1'
SUNNY_DAY_RECEIVER_PORT = base_port.next()
SUNNY_DAY_SENDER_PORT = base_port.next()


def test_sunny_day(assert_no_exceptions: Clearer,
                   wait_til_all_fit: Waiter) -> None:
    '''Train on bad data, recieve a message from a neighbor with good training,

    and check that we've improved.
    '''
    train_df, test_df, gt_df = load_classification_dataframes(reduced_train=True)  # noqa: F841

    pdef = ProblemDefinition(clause={
        'dataset': {
            'config': 'memory',
            'params': {
                'train_data': 'train_df',
                'test_data': 'test_df',
            },
            'column_roles': {
                'target': {
                    'name': 'target',
                    'pos_label': 0
                },
            },
        },
        'distributed': {
            'polling_interval': 0.5,
            'discoverer': {
                'name': 'static',
                'static': {
                    'adjacency': {
                        '1': [],
                    },
                },
            },
            'communicator': {
                'name': 'sockets',
                'sockets': {
                    'nodes_and_endpoints': [
                        ('1', (LOOPBACK, SUNNY_DAY_RECEIVER_PORT)),
                    ],
                },
            },
            'my_id': 1,
            'split': {
                'num_nodes': 2,
                'seed': 1234,
            },
            'pipelines': {
                'just': ['Auton_Logistic_Regression'],
            }
        },
        'problem_type': {
            'task': 'binary_classification',
        }
    })

    dut = DistributedWrangler(problem_definition=pdef)
    ground_truth = dut.dataset(data=gt_df, key='ground_truth')
    test_dataset = dut.dataset(data=test_df)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, SUNNY_DAY_SENDER_PORT))

    # These parameters were hand-extracted from a successful fit.
    message = pickle.dumps((
        1,
        np.array([5.72550884, 0.09525482, -0.89685941, 0.0]),
        None
    ))

    metric_catalog = MetricCatalogAuto()
    metric = metric_catalog.lookup_by_name('accuracy_score')
    try:
        dut.start()

        assert wait_til_all_fit(dut.all_distributed_instances, convergence_check=True), (
            'Failed to fit in time'
        )
        res1: Optional[PipelineResults] = dut.predict(test_dataset)
        assert res1 is not None
        assert len(res1) == 1

        # Confirm that the training is very bad.
        assert metric.calculate(pred=next(iter(res1.values())).prediction,
                                ground_truth=ground_truth
                                ) == pytest.approx(expected=0.06, abs=0.01)
        sock.sendto(message, (LOOPBACK, SUNNY_DAY_RECEIVER_PORT))
        # Wait until we have neighbor state.
        for d in dut.all_distributed_instances:
            while not list(d._neighbor_models_iter):  # pylint: disable=protected-access
                time.sleep(0.1)

        # Wait for the new state to get processed.
        assert wait_til_all_fit(
            dut.all_distributed_instances, min_time=1.0, convergence_check=True), (
            'Failed to fit in time for second time')

        res2 = dut.predict(test_dataset)
        assert res2 is not None
        assert len(res2) == 1

        # Confirm that the training is now better.
        assert metric.calculate(pred=next(iter(res2.values())).prediction,
                                ground_truth=ground_truth
                                ) >= 0.8
        assert_no_exceptions(dut.all_distributed_instances)
    finally:
        sock.close()
        dut.stop()


def test_lookup_templates() -> None:
    '''Test 3 ways of specifying pipeline loader args in the problem def.

    (single arg, multiple *args list, and **kwargs dict)
    '''
    pdef = ProblemDefinition(clause={
        'dataset': {
            'config': 'ignore'
        },
        'problem_type': {
            'task': 'test_task'
        },
        'distributed': {
            'communicator': {
                'name': 'stub_communicator'
            },
            'split': {
                'my_id': 1,
            },
            'pipelines': {
                'just': [
                    'auton_Logistic_Regression',
                    ['connect'],
                    {'alg': 'identity'}
                ]
            }
        }
    })
    dut = DistributedWrangler(problem_definition=pdef)
    dut.stop()
    got = dut.bound_pipelines
    want = {
        Designator('auton_logistic_regression'),
        Designator('connect'),
        Designator('identity'),
    }
    assert {pipe.designator for pipe in got.values()} == want


SUNNY_DAY_SERVER_MY_PORT = base_port.next()
SUNNY_DAY_SERVER_NEIGHBOR_PORT = base_port.next()
SUNNY_DAY_SERVER_WEB_PORT = base_port.next()


def test_server_sunny_day() -> None:
    '''Test the REST web server with a fit and predict request.'''
    train_df, test_df, gt_df = load_classification_dataframes(reduced_train=False)  # noqa: F841

    empty_df = pd.DataFrame()  # noqa: F841

    poll_queue: queue.Queue[pd.DataFrame] = queue.Queue()  # noqa: F841

    pdef = ProblemDefinition(clause={
        'dataset': {
            'config': 'ignore',
            'column_roles': {
                'target': {
                    'name': 'target',
                    'pos_label': 1,
                },
            },
        },
        'distributed': {
            'polling_interval': 0.5,
            'discoverer': {
                'name': 'static',
                'static': {
                    'adjacency': {
                        '1': ['2'],
                        '2': ['1'],
                    },
                },
            },
            'communicator': {
                'name': 'sockets',
                'sockets': {
                    'nodes_and_endpoints': [
                        ('1', (LOOPBACK, SUNNY_DAY_SERVER_MY_PORT)),
                        ('2', (LOOPBACK, SUNNY_DAY_SERVER_NEIGHBOR_PORT)),
                    ],
                },
            },
            'my_id': 1,
            'split': {
                'num_nodes': 2,
                'seed': 1234,
            },
            'pipelines': {
                'just': ['Auton_Logistic_regression'],
            }
        },
        'problem_type': {
            'task': 'binary_classification',
        }
    })

    dut = DistributedWrangler(problem_definition=pdef)

    train_msg = {
        'target': {'target': train_df['target'].tolist()},
        'covariates': train_df.drop(columns=['target']).to_dict(orient='list'),
    }

    test_msg = {
        'dataframe': test_df.to_dict(orient='list'),
    }

    # These parameters were hand-extracted from a successful fit.
    want_fit_message = (1, [-0.5086850644825086, 2.3477418500057716,
                        -0.26560121189885955, 19.2694391248473])

    # Create a fake neighbor to receive dut's state
    neighbor = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    neighbor.bind((LOOPBACK, SUNNY_DAY_SERVER_NEIGHBOR_PORT))
    neighbor.setblocking(False)
    server = None
    exception = None
    try:
        # Start web server.
        server = Process(target=dut.server(host=LOOPBACK, port=SUNNY_DAY_SERVER_WEB_PORT))
        server.start()

        time.sleep(2.0)

        # post a REST request to fit and provide data
        r = requests.post(f'http://localhost:{SUNNY_DAY_SERVER_WEB_PORT}/wrangler/v1.0/fit',
                          json={
                              'data': train_msg,
                          },
                          timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        assert r.text == '{"status":"UNTRAINED"}\n'
        # The dut should fit on the new data and send its state the neighbor.

        # Collect the state update from dut.
        ready = select.select([neighbor], [], [], 60.0)
        if ready[0]:
            neighbor_response = neighbor.recv(1024)
        else:
            raise TimeoutError('Timeout waiting for neighbor to respond')

        # Check that this looks like a properly fit state.
        response = pickle.loads(neighbor_response)
        assert response[1] == pytest.approx(want_fit_message[1], abs=0.01)

        # Ask dut to predict on test data.
        r = requests.post(f'http://localhost:{SUNNY_DAY_SERVER_WEB_PORT}/wrangler/v1.0/predict',
                          json={
                              'data': test_msg,
                          },
                          timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        got = r.json()

        assert 'auton_logistic_regression' in got

        # Check that the predictions are good.
        prediction = dut.dataset(
            pd.DataFrame(got['auton_logistic_regression']['predictions']),
            key='predictions')

        assert [str(k) for k in got['auton_logistic_regression']['predictions'].keys()] == [
            'target']
        assert [str(k) for k in got['auton_logistic_regression']['probabilities'].keys()] == [
            '0', '1']

        metric_catalog = MetricCatalogAuto()
        metric = metric_catalog.lookup_by_name('accuracy_score')
        ground_truth = dut.dataset(data=gt_df, key='ground_truth')

        assert metric.calculate(pred=prediction,
                                ground_truth=ground_truth) == pytest.approx(0.88, abs=0.01)

    except Exception as e:  # pylint: disable=broad-except
        exception = e
        log.error("exception %s in test_server_sunny_day", e)
    finally:
        if server is not None:
            server.terminate()
            server.join()
        neighbor.close()

    if exception is not None:
        raise exception


MY_ID_SERVER_MY_PORT = base_port.next()
MY_ID_SERVER_NEIGHBOR_PORT = base_port.next()
MY_ID_SERVER_WEB_PORT = base_port.next()


def test_server_set_my_id() -> None:
    '''Test starting the web server with a specified my_id.'''
    train_df, test_df, gt_df = load_classification_dataframes(reduced_train=True)  # noqa: F841

    pdef = ProblemDefinition(clause={
        'dataset': {
            'config': 'ignore',
            'column_roles': {
                'target': {
                    'name': 'target',
                    'pos_label': 1,
                },
            },
        },
        'distributed': {
            'polling_interval': 0.5,
            'discoverer': {
                'name': 'static',
                'static': {
                    'adjacency': {
                        '1': ['2'],
                        '2': ['1'],
                    },
                },
            },
            'communicator': {
                'name': 'sockets',
                'sockets': {
                    'nodes_and_endpoints': [
                        ('1', (LOOPBACK, MY_ID_SERVER_MY_PORT)),
                        ('2', (LOOPBACK, MY_ID_SERVER_NEIGHBOR_PORT)),
                    ],
                },
            },
            'my_id': 0,
            'split': {
                'num_nodes': 2,
                'seed': 1234,
            },
            'pipelines': {
                'just': ['Auton_Logistic_regression'],
            }
        },
        'problem_type': {
            'task': 'binary_classification',
        }
    })

    dut = DistributedWrangler(problem_definition=pdef, my_id=1)

    train_msg = {
        'target': {'target': train_df['target'].tolist()},
        'covariates': train_df.drop(columns=['target']).to_dict(orient='list'),
    }

    neighbor = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    neighbor.bind((LOOPBACK, MY_ID_SERVER_NEIGHBOR_PORT))
    neighbor.setblocking(False)
    server = None
    exception = None
    try:
        server = Process(target=dut.server(host=LOOPBACK, port=MY_ID_SERVER_WEB_PORT))
        server.start()

        time.sleep(2.0)

        r = requests.post(f'http://localhost:{MY_ID_SERVER_WEB_PORT}/wrangler/v1.0/fit',
                          json={
                              'data': train_msg,
                          },
                          timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        assert r.text == '{"status":"UNTRAINED"}\n'

        ready = select.select([neighbor], [], [], 60.0)
        if ready[0]:
            _, response_addr = neighbor.recvfrom(1024)
        else:
            raise TimeoutError('Timeout waiting for neighbor to respond')

        assert response_addr == (LOOPBACK, MY_ID_SERVER_MY_PORT)

    except Exception as e:  # pylint: disable=broad-except
        exception = e
        log.error("exception %s in test_server_sunny_day", e)
    finally:
        if server is not None:
            server.terminate()
            server.join()

    if exception is not None:
        raise exception


SERVER_METRICS_MY_PORT = base_port.next()
SERVER_METRICS_WEB_PORT = base_port.next()


def test_server_predict_and_score() -> None:
    '''Test the REST web server with a request to evaluate metrics.'''
    train_df, test_df, gt_df = load_classification_dataframes(reduced_train=False)  # noqa: F841

    pdef = ProblemDefinition(clause={
        'dataset': {
            'config': 'ignore',
            'column_roles': {
                'target': {
                    'name': 'target',
                    'pos_label': 1,
                },
            },
        },
        'distributed': {
            'polling_interval': 0.5,
            'discoverer': {
                'name': 'static',
                'static': {
                    'adjacency': {
                        '1': [],
                    },
                },
            },
            'communicator': {
                'name': 'sockets',
                'sockets': {
                    'nodes_and_endpoints': [
                        ('1', (LOOPBACK, SERVER_METRICS_MY_PORT)),
                    ],
                },
            },
            'my_id': 1,
            'split': {
                'num_nodes': 2,
                'seed': 1234,
            },
            'pipelines': {
                'just': ['Auton_Logistic_regression'],
            }
        },
        'problem_type': {
            'task': 'binary_classification',
        }
    })

    dut = DistributedWrangler(problem_definition=pdef)

    train_msg = {
        'target': {'target': train_df['target'].tolist()},
        'covariates': train_df.drop(columns=['target']).to_dict(orient='list'),
    }

    test_msg = {
        'dataframe': test_df.to_dict(orient='list'),
        'ground_truth': gt_df.to_dict(orient='list')
    }

    server = None
    exception = None
    try:
        # Start web server.
        server = Process(target=dut.server(host=LOOPBACK, port=SERVER_METRICS_WEB_PORT))
        server.start()

        time.sleep(2.0)

        # post a REST request to fit and provide data
        r = requests.post(f'http://localhost:{SERVER_METRICS_WEB_PORT}/wrangler/v1.0/fit',
                          json={
                              'data': train_msg,

                          },
                          timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        assert r.text == '{"status":"UNTRAINED"}\n'
        # The dut should fit on the new data and send its state to the neighbor.
        r = requests.get(
            f'http://localhost:{SERVER_METRICS_WEB_PORT}/wrangler/v1.0/wait_til_all_fit',
            timeout=30)
        assert r.ok, (
            f'Response code to wait_til_all_fit was {r}'
        )

        # Ask dut to predict on test data.
        r = requests.post(
            f'http://localhost:{SERVER_METRICS_WEB_PORT}/wrangler/v1.0/predict_and_score',
            json={
                'data': test_msg,
                'metrics': ['accuracy_score'],
            },
            timeout=None)
        assert r.ok, (
            f'Response code to predict_and_score was {r}'
        )

        got = r.json()

        # want = {
        #     'auton_logistic_regression': {
        #         'predictions': ...,
        #         'metrics': {
        #             'accuracy_score': 0.88
        #         }
        #     },
        # }

        assert list(got.keys()) == ['auton_logistic_regression']
        assert got['auton_logistic_regression']['metrics']['accuracy_score'] == pytest.approx(
            0.88, rel=0.0001
        )
        assert 'predictions' in got['auton_logistic_regression']
        assert 'probabilities' in got['auton_logistic_regression']
        assert [str(k) for k in got['auton_logistic_regression']['predictions'].keys()] == [
            'target']
        assert [str(k) for k in got['auton_logistic_regression']['probabilities'].keys()] == [
            '0', '1']

    except Exception as e:  # pylint: disable=broad-except
        exception = e
        log.error("exception %s in test_server_predict_and_score", e)
    finally:
        if server is not None:
            server.terminate()
            server.join()

    if exception is not None:
        raise exception


def test_2_wranglers(assert_no_exceptions: Clearer) -> None:
    '''2 distributed wranglers are given the same data, we want them to converge with no errors.'''
    breast_cancer = datasets.load_breast_cancer(as_frame=True)
    df = breast_cancer.frame

    train = df.iloc[0:-100,]  # noqa: F841 pylint: disable=unused-variable
    test = df.tail(50)
    del test['target']
    loopback = '127.0.0.1'

    pdef_dict_1 = {
        "dataset": {
            "config": "memory",
            "params": {
                "train_data": "train",
                "test_data": "test"
            },
            "column_roles": {
                "target": {
                    "name": "target",
                    "pos_label": 1
                }
            }
        },
        "problem_type": {
            "data_type": "tabular",
            "task": "binary_classification"
        },
        "metrics": {
            "accuracy_score": {},
            "roc_auc_score": {}
        },
        'distributed': {
            'polling_interval': 0.5,
            'fit_eps': 0.5,
            'discoverer': {
                'name': 'static',
                'static': {
                    'adjacency': {
                        '1': [2],
                        '2': [1]
                    },
                },
            },
            'communicator': {
                'name': 'sockets',
                'sockets': {
                    'nodes_and_endpoints': [
                        ('1', (loopback, 65001)),
                        ('2', (loopback, 65002)),
                    ],
                },
            },
            'my_id': 1,
            'split': {
                'num_nodes': 2
            },
            'pipelines': {
                'just': ['Auton_Logistic_Regression'],
            }
        },
        'hyperparams': [
            {
                '_comments': [
                    'Apply standard parameters to instances of Auton_Logistic_Regression.',
                    'Note that algorithm is the name of the algorithm associated with',
                    'the step where name is name of the step.'
                ],
                'select': {
                    'algorithm': 'Auton_Logistic_Regression'
                },
                'params': {
                    'L2': {
                        'fixed': 5.0
                    },
                    'Lambda': {
                        'fixed': 10000000
                    },
                    'omega': {
                        'fixed': 0.6
                    },
                    'synchronous': {
                        'fixed': True
                    },
                },
            }
        ]
    }

    # The second node's problem def is identical aside from my_id
    pdef_dict_2 = deepcopy(pdef_dict_1)
    dist_clause = pdef_dict_2['distributed']
    assert isinstance(dist_clause, dict)
    dist_clause['my_id'] = 2

    pdef_1 = ProblemDefinition(pdef_dict_1)
    pdef_2 = ProblemDefinition(pdef_dict_2)

    dut1 = DistributedWrangler(problem_definition=pdef_1)
    dut2 = DistributedWrangler(problem_definition=pdef_2)

    try:

        # fit on local train data from the problem def and start listening for messages
        dut1.start()
        dut2.start()

        # assert that we converge
        all_instances = list(dut1.all_distributed_instances) + list(dut2.all_distributed_instances)
        for _ in range(5):
            advance(all_instances)

        assert_no_exceptions(all_instances)

    finally:
        dut1.stop()
        dut2.stop()


SERVER_PROBA_MY_PORT = base_port.next()
SERVER_PROBA_WEB_PORT = base_port.next()


def test_server_proba() -> None:
    '''Test the REST web server returning probabilities for classification.'''
    train_df, test_df, gt_df = load_classification_dataframes(reduced_train=False)  # noqa: F841

    pdef = ProblemDefinition(clause={
        'dataset': {
            'config': 'ignore',
            'column_roles': {
                'target': {
                    'name': 'target',
                    'pos_label': 1,
                },
            },
        },
        'distributed': {
            'polling_interval': 0.5,
            'discoverer': {
                'name': 'static',
                'static': {
                    'adjacency': {
                        '1': [],
                    },
                },
            },
            'communicator': {
                'name': 'sockets',
                'sockets': {
                    'nodes_and_endpoints': [
                        ('1', (LOOPBACK, SERVER_PROBA_MY_PORT)),
                    ],
                },
            },
            'my_id': 1,
            'pipelines': {
                'just': ['Auton_Decision_Tree_Model'],
            }
        },
        'problem_type': {
            'task': 'binary_classification',
        }
    })

    dut = DistributedWrangler(problem_definition=pdef)

    train_msg = {
        'target': {'target': train_df['target'].tolist()},
        'covariates': train_df.drop(columns=['target']).to_dict(orient='list'),
    }

    test_msg = {
        'dataframe': test_df.to_dict(orient='list'),
        'ground_truth': gt_df.to_dict(orient='list')
    }

    server = None
    exception = None
    try:
        # Start web server.
        server = Process(target=dut.server(host=LOOPBACK, port=SERVER_PROBA_WEB_PORT))
        server.start()

        time.sleep(2.0)

        # post a REST request to fit and provide data
        r = requests.post(f'http://localhost:{SERVER_PROBA_WEB_PORT}/wrangler/v1.0/fit',
                          json={
                              'data': train_msg,

                          },
                          timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        assert r.text == '{"status":"UNTRAINED"}\n'
        # The dut should fit on the new data and send its state the neighbor.

        r = requests.get(f'http://localhost:{SERVER_PROBA_WEB_PORT}/wrangler/v1.0/wait_til_all_fit',
                         timeout=20)
        assert r.ok, (
            f'Response code to wait_til_all_fit was {r}'
        )

        # Post a REST request to predict, as well as predict_and_score (we want to test both.)
        gots = []

        r = requests.post(
            f'http://localhost:{SERVER_PROBA_WEB_PORT}/wrangler/v1.0/predict',
            json={
                'data': test_msg,
            },
            timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        gots.append(r.json())

        r = requests.post(
            f'http://localhost:{SERVER_PROBA_WEB_PORT}/wrangler/v1.0/predict_and_score',
            json={
                'data': test_msg,
                'metrics': ['accuracy_score'],
            },
            timeout=None)
        assert r.ok, (
            f'Response code to predict_and_score was {r}'
        )

        gots.append(r.json())

        for i, got in enumerate(gots):
            assert list(got.keys()) == ['auton_decision_tree_model'], (
                f'Error in got[{i}]: {got}'
            )
            assert 'probabilities' in got['auton_decision_tree_model']
            df = pd.DataFrame(got['auton_decision_tree_model']['probabilities'])
            assert df.shape == (test_df.shape[0], 2)
            assert list(df.columns) == ['0', '1']
            # If these are class probabilities, every row should sum to 1
            #       (Accounting for float imprecision)
            assert all(s == pytest.approx(1.0) for s in (df['0'] + df['1']))

    except Exception as e:  # pylint: disable=broad-except
        exception = e
        log.error("exception %s in test_server_proba", e)
    finally:
        if server is not None:
            server.terminate()
            server.join()

    if exception is not None:
        raise exception


def test_server_mean() -> None:
    '''Test that the REST server works with AutonMean'''
    train_df = pd.DataFrame(
        {
            'a': [None, None, 1],
            'b': [None, 2, 3],
            'c': [4, 5, 6]
        }
    )

    pdef = ProblemDefinition(clause={
        'dataset': {
            'config': 'ignore',
            'column_roles': {
                'target': {
                    'name': 'a',
                },
            },
        },
        'distributed': {
            'polling_interval': 0.5,
            'discoverer': {
                'name': 'static',
                'static': {
                    'adjacency': {
                        '1': [],
                    },
                },
            },
            'communicator': {
                'name': 'sockets',
                'sockets': {
                    'nodes_and_endpoints': [
                        ('1', (LOOPBACK, SERVER_PROBA_MY_PORT)),
                    ],
                },
            },
            'my_id': 1,
            'pipelines': {
                'just': ['Auton_Mean'],
            }
        },
        'problem_type': {
            'task': 'other',
        }
    })

    dut = DistributedWrangler(problem_definition=pdef)

    msg = {
        'data': {
            'dataframe': df_to_dict_json_friendly(train_df)
        }
    }

    server = None
    exception = None
    try:
        # Start web server.
        server = Process(target=dut.server(host=LOOPBACK, port=SERVER_PROBA_WEB_PORT))
        server.start()

        time.sleep(2.0)

        # post a REST request to fit and provide data
        r = requests.post(f'http://localhost:{SERVER_PROBA_WEB_PORT}/wrangler/v1.0/fit',
                          json=msg,
                          timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        assert r.text == '{"status":"UNTRAINED"}\n'

        # The dut should fit on the new data and send its state to the neighbor.
        r = requests.get(f'http://localhost:{SERVER_PROBA_WEB_PORT}/wrangler/v1.0/wait_til_all_fit',
                         timeout=20)
        assert r.ok, (
            f'Response code to wait_til_all_fit was {r}'
        )

        # Post a REST request to predict to get the mean
        r = requests.post(
            f'http://localhost:{SERVER_PROBA_WEB_PORT}/wrangler/v1.0/predict',
            json=msg,
            timeout=None)
        assert r.ok, (
            f'Response code to predict was {r}'
        )
        got = r.json()
        assert list(got.keys()) == ['auton_mean'], (
            f'Error in got: {got}'
        )
        pred_df = pd.DataFrame(got['auton_mean']['predictions'])

        pd.testing.assert_frame_equal(
            pred_df,
            pd.DataFrame({'a': [1.0], 'b': [2.5], 'c': [5.0]}))

    except Exception as e:  # pylint: disable=broad-except
        exception = e
        log.error("exception %s in test_server_mean", e)
    finally:
        if server is not None:
            server.terminate()
            server.join()

    if exception is not None:
        raise exception



'''The "simple" executor'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import logging
from typing import Dict, Optional
import traceback

from .simple_executable_pipeline import SimpleExecutablePipeline
from ..executor import Executor, ExecutorKind, TrainedPipelineCollection
from ...generator.designator import Designator
from ...instantiator.executable_pipeline import (ExecutablePipeline, PipelineResult,
                                                 FitError, PredictError)
from ...wrangler.dataset import Dataset, DatasetKeys


class SimpleExecutor(Executor):
    '''A simple implementation of an Executor'''
    _kind = ExecutorKind('simple')

    def fit(self,
            dataset: Optional[Dataset],
            pipelines: Dict[Designator, ExecutablePipeline]
            ) -> TrainedPipelineCollection:
        '''Train a list of pipelines on a dataset.'''
        retval = TrainedPipelineCollection()
        for des, pipe in pipelines.items():
            assert isinstance(pipe, SimpleExecutablePipeline)
            try:
                retval.set(des, pipe.fit(dataset=dataset))
            except Exception:  # pylint: disable=broad-exception-caught
                trace = traceback.format_exc()
                logging.warning(
                    'Encountered error in the %s pipeline during fit:\n%s',
                    des,
                    trace)
                pipe.set_fit_error(FitError(trace))
                retval.set(des, FitError(trace))
        return retval

    def predict(self,
                dataset: Dataset,
                pipelines: Dict[Designator, ExecutablePipeline]
                ) -> Dict[Designator, PipelineResult]:
        '''Use a list of simple pipelines to predict from a dataset.'''
        retval: Dict[Designator, PipelineResult] = {}
        for des, pipe in pipelines.items():
            assert isinstance(pipe, SimpleExecutablePipeline)
            try:
                retval[des] = pipe.predict(dataset=dataset)
            except Exception:  # pylint: disable=broad-exception-caught
                trace = traceback.format_exc()
                logging.warning(
                    'Encountered error in the %s pipeline during predict:\n%s',
                    des,
                    trace)
                out_data = dataset.output()
                out_data[DatasetKeys.ERROR.value] = PredictError(trace)
                retval[des] = PipelineResult(
                    prediction=out_data,
                    executable_pipeline=pipe)
        return retval



'''Test the pipeline template catalog module.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code

import pytest

from .pipeline_template import PipelineTemplate
from .template import TemplateCatalog, TemplateRegisterError


def test_register() -> None:
    '''Test registering a template'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        name='TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    assert dut.register(test_template) == 'testtemplate'


def test_register_with_strings() -> None:
    '''Test registering a template with strings instead of lists in tags'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        'TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    assert dut.register(test_template) == 'testtemplate'
    assert dut.lookup_by_task(task='some task') == {'testtemplate': test_template}


def test_register_no_task() -> None:
    '''Attempt to register template without task'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        'TestTemplate',
        tags={'data_type': ['Some Type'], 'task': []})
    dut.register(test_template)
    with pytest.raises(TemplateRegisterError,
                       match=r"[Tt]ask"):
        dut.validate()


def test_register_no_datatype() -> None:
    '''Attempt to register template without datatype'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        'TestTemplate',
        tags={'data_type': [], 'task': ['Some Task']})
    dut.register(test_template)
    with pytest.raises(TemplateRegisterError,
                       match=r"[Dd]ata"):
        dut.validate()


def test_register_fails() -> None:
    '''Attempt to register template without datatype.

    Expected behavior: a TemplateRegisterError will be thrown.
    (The template may still register)
    '''
    dut = TemplateCatalog()
    test_template_bad = PipelineTemplate(
        'TestTemplateBad',
        tags={'data_type': [], 'task': ['Another Task']})
    dut.register(test_template_bad)
    with pytest.raises(TemplateRegisterError,
                       match=r"[Dd]ata"):
        dut.validate()


def test_lookup_by_task() -> None:
    '''Test finding a template by task'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        'TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    assert 1 == len(dut.lookup_by_task(['Some Task']))


def test_lookup_by_task_fail() -> None:
    '''Test looking up a nonexistent task'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        'TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    assert 0 == len(dut.lookup_by_task('Nonexistent'))


def test_lookup_by_task_multi() -> None:
    '''Test finding multiple templates by task'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        'TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    test_second = PipelineTemplate(
        'TestSecond',
        tags={'data_type': ['Another Type'], 'task': ['Some Task']})
    dut.register(test_second)
    assert 2 == len(dut.lookup_by_task('Some Task'))


def test_lookup_by_datatype() -> None:
    '''Test finding a template by datatype'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        'TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    assert 1 == len(dut.lookup_by_datatype('Some Type'))


def test_lookup_by_datatype_fail() -> None:
    '''Test looking up a nonexistent datatype'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        name='TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    assert 0 == len(dut.lookup_by_datatype('Nonexistent'))


def test_lookup_by_datatype_multi() -> None:
    '''Test finding multiple templates by datatype'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        name='TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    test_second = PipelineTemplate(
        name='TestSecond',
        tags={'data_type': ['Some Type'], 'task': ['Another Task']})
    dut.register(test_second)
    assert 2 == len(dut.lookup_by_datatype('Some Type'))


def test_lookup_by_task_and_type() -> None:
    '''Test finding a template by task and datatype'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        name='TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    assert 1 == len(dut.lookup_by_both(task='Some Task',
                                       data_type='Some Type'))


def test_lookup_by_task_and_type_fail() -> None:
    '''Test looking up a nonexistent task and datatype'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        name='TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    assert 0 == len(dut.lookup_by_both(task='Nonexistent',
                                       data_type='Nonexistent'))


def test_lookup_by_task_and_type_mismatch() -> None:
    '''Test looking up a task but datatype does not match'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        name='TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    assert 0 == len(dut.lookup_by_both(task='Some Task',
                                       data_type='Nonexistent'))


def test_lookup_by_task_and_type_multi() -> None:
    '''Test finding multiple templates by task and datatype'''
    dut = TemplateCatalog()
    test_template = PipelineTemplate(
        name='TestTemplate',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_template)
    test_second = PipelineTemplate(
        name='TestSecond',
        tags={'data_type': ['Some Type'], 'task': ['Some Task']})
    dut.register(test_second)
    test_third = PipelineTemplate(
        name='TestThird',
        tags={'data_type': ['Some Type'], 'task': ['Different Task']})
    dut.register(test_third)
    assert 2 == len(dut.lookup_by_both(task='Some Task',
                                       data_type='Some Type'))



'''A Table representing a pandas Series.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import List, Union

import numpy as np
import pandas as pd

from .impl.table import TableFactory, Table
from .impl.table_catalog import TableCatalog

from .dataframe import DataFrameTable


class SeriesTable(DataFrameTable):
    '''A Table representing a pandas Series.

    This Table can only have one column.
    '''
    name = 'series'
    _value: pd.Series  # type: ignore[assignment]
    my_type = pd.Series  # type: ignore[assignment]

    @property
    def columns(self) -> List[Union[int, str]]:
        '''Return the column names of the table.'''
        return [str(self._value.name)]

    @columns.setter
    def columns(self, value: List[Union[int, str]]) -> None:
        '''Set the column names of the table.'''
        self._value.name = value[0]


def as_numpy(from_table: Table) -> np.ndarray:
    '''Export a SeriesTable to a numpy array.'''
    assert isinstance(from_table, SeriesTable)
    if from_table.shape[1] == 1:
        return from_table.value().to_numpy().flatten()
    return from_table.value().to_numpy()


def as_dataframe(from_table: Table) -> pd.DataFrame:
    '''Export a SeriesTable to a pandas DataFrame.'''
    assert isinstance(from_table, SeriesTable)
    return pd.DataFrame(from_table.value())


def register(catalog: TableCatalog):
    '''Register the SeriesTable with the catalog.'''
    catalog.register(SeriesTable)
    TableFactory.register_constructor(pd.Series, SeriesTable)
    TableFactory.register_exporter(
        from_type=pd.Series, to_type=pd.Series, exporter=SeriesTable.value)
    TableFactory.register_exporter(
        from_type=pd.Series, to_type=np.ndarray, exporter=as_numpy)
    TableFactory.register_exporter(
        from_type=pd.Series, to_type=pd.DataFrame, exporter=as_dataframe)



'''Tests for ParallelStep'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any

from ...algorithms.impl.algorithm import Algorithm, AlgorithmCatalogStub

from .query_step import QueryStep
from .pipeline_template import PipelineTemplate

# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code,protected-access


class ModelStub(Algorithm):
    def instantiate(self, **unused_kwargs) -> Any:
        return None


def test_query() -> None:
    model = ModelStub()
    algorithm_catalog = AlgorithmCatalogStub()
    algorithm_catalog.register(model, name='Some_Model', tags={'some_tag': ['some_value']})

    dut = PipelineTemplate(name='Fake_Template', algorithm_catalog=algorithm_catalog)
    query_step = dut.query(some_tag='some_value')
    assert len(dut._steps) == 1
    assert isinstance(query_step, QueryStep)



'''Tests for catalog_element_mixin.py'''
# pylint: disable=missing-function-docstring
import pytest

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .memory_catalog import CatalogNameError
from .catalog_element_mixin import CatalogElementMixin

# pylint: disable=too-few-public-methods,missing-function-docstring,missing-class-docstring,duplicate-code


class Widget(CatalogElementMixin):
    '''This is the thing that is stored in WidgetCatalog.'''


class WidgetNameAndTags(Widget):
    _name = 'widget_with_name_and_tags'
    _tags = {
        'widget_tag': ['widget_value'],
    }


class WidgetUppercase(Widget):
    _name = 'Widget_with_Uppercase'
    _tags = {
        'Widget_Tag': ['Widget_Value'],
    }


class WidgetNameNoTags(Widget):
    _name = 'widget_with_name'


def test_sunny_day():
    dut = WidgetNameAndTags()
    assert dut.name == 'widget_with_name_and_tags'
    assert dut.tags['widget_tag'] == ['widget_value']


def test_widget_name_no_tags():
    dut = WidgetNameNoTags()
    assert dut.name == 'widget_with_name'
    assert not dut.tags


def test_missing_name_should_fail():
    with pytest.raises(CatalogNameError, match='Widget'):
        Widget()


def test_uppercase():
    '''Confirm that name and tags are all forced to lower case.'''
    dut = WidgetUppercase()
    assert dut.name == 'widget_with_uppercase'
    assert dut.tags['widget_tag'] == ['widget_value']



'''Tests for auton_simple_domain_confusion.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
import numpy as np
import pytest

from ...wrangler.dataset import Dataset

from .auton_moment_matcher import AutonMomentMatcher
from ..moments import Moments

# pylint: disable=missing-function-docstring,duplicate-code


def test_identity():

    source = Dataset(
        dataframe=pd.DataFrame({
            'a': [1.5, 2.5, 3.5],
            'b': [4.0, 5.0, 6.0]
        }),
    )
    preprocessor = Moments().instantiate()
    preprocessor.fit(source)
    source_params = preprocessor.predict(None)

    train_population = Dataset(
        dataframe=pd.DataFrame({
            'a': [1.5, 2.5, 3.5],
            'b': [4.0, 5.0, 6.0]
        }),
        hyperparams=source_params['hyperparams']
    )

    test_population = Dataset(
        dataframe=pd.DataFrame({
            'a': [1.5, 2.5, 3.5],
            'b': [4.0, 5.0, 6.0]
        }),
        hyperparams=source_params['hyperparams']
    )

    model = AutonMomentMatcher()
    dut = model.instantiate()

    dut.fit(train_population)

    got = dut.predict(test_population)

    np.testing.assert_allclose(got.dataframe.to_numpy(),
                               test_population.dataframe.to_numpy(), atol=.001)


def test_mean_shifted():

    source = Dataset(
        dataframe=pd.DataFrame({
            'a': [3.5, 1.5, 7.5],
            'b': [4.0, 5.0, 6.0]
        }),
    )
    preprocessor = Moments().instantiate()
    preprocessor.fit(source)
    source_params = preprocessor.predict(None)

    train_population = Dataset(
        dataframe=pd.DataFrame({
            'a': np.array([3.5, 1.5, 7.5]) + 10,
            'b': np.array([4.0, 5.0, 6.0]) + 10

        }),
        hyperparams=source_params['hyperparams']
    )

    model = AutonMomentMatcher()
    dut = model.instantiate()

    dut.fit(train_population)

    got = dut.predict(train_population)

    np.testing.assert_allclose(got.dataframe.mean(axis=0),
                               np.array([np.mean([3.5, 1.5, 7.5]), 5]), atol=.001)


def test_sunny_day():

    source_cov = np.array([[1, 2], [-3, 9]])
    source_mu = np.array([-6, 3])
    target_cov = np.array([[8, -1], [1, 4]])
    target_mu = np.array([0.2, 7.3])

    # Ensure covariances are PSD
    source_cov = source_cov.T @ source_cov
    target_cov = target_cov.T @ target_cov

    np.random.seed(1)
    source_df = pd.DataFrame(
        data=np.random.multivariate_normal(source_mu, source_cov, size=10000),
        columns=['a', 'b'])
    target_train_df = pd.DataFrame(
        data=np.random.multivariate_normal(target_mu, target_cov, size=10000),
        columns=['a', 'b'])
    target_test_df = pd.DataFrame(
        data=np.random.multivariate_normal(target_mu, target_cov, size=10000),
        columns=['a', 'b'])

    source = Dataset(dataframe=source_df)
    preprocessor = Moments().instantiate()
    preprocessor.fit(source)
    source_params = preprocessor.predict(None)

    train_population = Dataset(
        dataframe=target_train_df,
        hyperparams=source_params['hyperparams']
    )

    test_population = Dataset(
        dataframe=target_test_df
    )

    model = AutonMomentMatcher()
    dut = model.instantiate()

    dut.fit(train_population)

    got = dut.predict(test_population)

    np.testing.assert_allclose(got.dataframe.mean(axis=0), source_mu, rtol=.05)
    np.testing.assert_allclose(np.cov(got.dataframe.T), source_cov, rtol=.05)


def test_hyperparams_required():
    model = AutonMomentMatcher()
    dut = model.instantiate()

    with pytest.raises(ValueError,
                       match=r'mean.*required.*confusion.*\n.*covariance.*required.*confusion'):
        dut.hyperparams()


def test_hyperparams_set_in_constructor():
    model = AutonMomentMatcher()
    dut = model.instantiate(mean=np.array([1, 2]), covariance=np.array([[1, 2], [3, 4]]))

    assert np.all(dut.hyperparams()['mean'] == np.array([1, 2]))
    assert np.all(dut.hyperparams()['covariance'] == np.array([[1, 2], [3, 4]]))


def test_hyperparams_override():
    model = AutonMomentMatcher()
    base_mean = np.array([3, 4])
    base_covariance = np.array([[3, 4], [5, 6]])
    override_covariance = np.array([[1, 2], [3, 4]])
    dut = model.instantiate(mean=base_mean, covariance=base_covariance)

    got = dut.hyperparams(covariance=override_covariance)
    assert np.all(got['mean'] == base_mean)
    assert np.all(got['covariance'] == override_covariance)



'''Distributed Logistic Regression'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# Linear and Logistic Regression have portions of identical code that should
# not be abstracted away.
# pylint complains about various members of autograd.numpy being missing,
# but we see them when we check manually.
# pylint: disable=duplicate-code,no-member

import pickle
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import autograd.numpy as np  # type: ignore[import]
from autograd import jacobian  # type: ignore[import]
from autograd.scipy.special import expit  # type: ignore[import]

from ...algorithms.impl.distributed_algorithm_instance import NoDataError  # type: ignore[import]
from ...catalog.catalog import upcast
from ...neighbor_manager.node_id import NodeID
from ...problem_def.task import DataType, TaskType
from ...wrangler.dataset import Dataset, DatasetKeys, RoleName, MetadataError
from ...wrangler.logger import Logger

from ..impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..impl.linear_distributed_algorithm_instance import (
    LinearDistributedAlgorithmInstance, LinearNeighborState,
    norm2, inner_product)


log = Logger(__file__).logger()


def _validate_xv(x: np.ndarray, v: np.ndarray) -> None:
    '''Validate the dimensions of x and v.

    x: np.ndarray with 2 or fewer dimensions
    y: np.ndarray with 1 dimension
    '''
    assert x.ndim <= 2, (
        f'BUG: exepcted x.ndim <=2, instead found {x.ndim}'
    )
    assert v.ndim == 1, (
        'BUG: expected 1 dimensional v; instead found {v.ndim}'
    )


def _grad_log_phi(x: np.ndarray, v: np.ndarray) -> np.ndarray:
    '''Return the gradient of the log of phi(x, v).'''
    _validate_xv(x, v)
    eps = 1e-15
    phi = _phi(x, v)
    if isinstance(phi, np.ndarray):
        phi = phi[:, np.newaxis]
    return _grad_phi(x, v) / (phi + eps)


def _grad_log_1mphi(x: np.ndarray, v: np.ndarray) -> np.ndarray:
    '''Return the gradient of the log of 1 - phi(x, v).'''
    _validate_xv(x, v)
    eps = 1e-15
    phi = _phi(x, v)
    if isinstance(phi, np.ndarray):
        phi = phi[:, np.newaxis]
    return -_grad_phi(x, v) / (1 - phi + eps)


def _grad_phi(x: np.ndarray, v: np.ndarray) -> np.ndarray:
    '''Return the gradient of the logistic function of x and v.'''
    _validate_xv(x, v)
    xv = x @ v
    if x.ndim > 1:
        return (np.exp(-xv) / (1.0 + np.exp(-xv)) ** 2)[:, np.newaxis] * x
    return (np.exp(-xv) / (1.0 + np.exp(-xv)) ** 2) * x


def _log_phi(x: np.ndarray, v: np.ndarray) -> Union[np.float64, np.ndarray]:
    '''Return the log of phi(x, v).'''
    _validate_xv(x, v)
    eps = 1e-15
    return np.log(_phi(x, v) + eps)


def _log_1mphi(x: np.ndarray, v: np.ndarray) -> Union[np.float64, np.ndarray]:
    '''Return the log of 1 - phi(x, v).'''
    _validate_xv(x, v)
    eps = 1e-15
    return np.log(1 - _phi(x, v) + eps)


def _phi(x: np.ndarray, v: np.ndarray) -> Union[np.float64, np.ndarray]:
    '''Return the logistic function of x and v.'''
    _validate_xv(x, v)
    # Fixes overflow error (return 1.0 / (1.0 + np.exp(-x @ v)))
    return expit(x @ v)


class AutonLogisticRegressionNeighbor(LinearNeighborState):
    '''A neighbor in the distributed logistic regression model.'''
    _positive_class: int

    def __init__(self, positive_class: int, v: np.ndarray, columns: Optional[List[str]]):
        super().__init__(v=v, columns=columns)
        assert isinstance(positive_class, int)  # mypy does not catch when this is violated
        self._positive_class = positive_class

    def encode(self) -> bytes:
        '''Encode message for distributed neighbors.'''
        #  minimal: communicate as little information as possible
        return pickle.dumps(
            (self._positive_class, self._v, self.columns))

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'AutonLogisticRegressionNeighbor':
        '''Decode a message from a neighbor.'''
        # minimal: communicate as little information as possible
        positive_class, v, cols = pickle.loads(serialized_model)
        assert isinstance(positive_class, int), (
            'BUG: expected positive_class to be an int, '
            f'instead found {positive_class} of type {type(positive_class)}'
        )
        assert isinstance(v, np.ndarray), (
            f'BUG: expected v to be an np.ndarray, instead found {v} of type {type(v)}'
        )
        if cols is not None:
            cols = cls._cast_columns(cols)
        return cls(positive_class=positive_class,
                   v=v,
                   columns=cols)

    @property
    def positive_class(self) -> int:
        '''The numeric identifier of the positive class in the classification problem.'''
        return self._positive_class

    @positive_class.setter
    def positive_class(self, value: int) -> None:
        '''The numeric identifier of the positive class in the classification problem.'''
        assert isinstance(value, int)  # mypy does not catch when this is violated.
        self._positive_class = value


class AutonLogisticRegressionInstance(LinearDistributedAlgorithmInstance):
    '''Binary logistic regression that supports distributed AI.

    As a distributed model, this instance can share information about its
    trained state with other instances and update its trained state using
    information shared by other instances.
    '''
    _neighbor_constructor = AutonLogisticRegressionNeighbor

    # This is a type annotation for _my_state.
    _predict_state: Optional[AutonLogisticRegressionNeighbor] = None

    # _my_state: Optional[AutonLogisticRegressionNeighbor] = None
    @property  # type: ignore[override]
    def _my_state(self) -> Optional[AutonLogisticRegressionNeighbor]:
        retval = LinearDistributedAlgorithmInstance._my_state.fget(self)  # type: ignore[attr-defined] # pylint: disable=assignment-from-no-return, line-too-long
        assert retval is None or isinstance(retval, AutonLogisticRegressionNeighbor), (
            'BUG: expected _my_state to be None or an AutonLogisticRegressionNeighbor.')
        return retval

    @_my_state.setter
    def _my_state(self, value: Optional[AutonLogisticRegressionNeighbor]) -> None:
        assert value is None or isinstance(value, AutonLogisticRegressionNeighbor), (
            'BUG: expected value to be None or an AutonLogisticRegressionNeighbor.')
        LinearDistributedAlgorithmInstance._my_state.fset(self, value)  # type: ignore[attr-defined]

    @property
    def _yhat_is_proba(self) -> bool:
        return True

    @property
    def _neighbor_models_iter(self) -> Iterable[AutonLogisticRegressionNeighbor]:
        for v in super()._neighbor_models_iter:
            assert isinstance(v, AutonLogisticRegressionNeighbor), (
                'BUG: expected neighbor_models to contain AutonLogisticRegressionNeighbor, '
                f'instead found {v} of type {type(v)}'
            )
            yield v

    def _standardize_y(self, y: np.ndarray) -> np.ndarray:
        '''Regularize the target values.'''
        positive_class = 1
        if self._my_state is not None and self._my_state.positive_class is not None:
            positive_class = self._my_state.positive_class
        return np.array(
            [1 if y_ == positive_class else 0 for y_ in y])

    def _yhat(self, x: np.ndarray) -> np.ndarray:
        assert self._predict_state is not None, (
            'BUG: self._predict_state should not be None when _yhat is called.'
            ' This should have been set in _fit'
        )
        yhat = np.empty((x.shape[0], 2))
        yhat[:, 1] = _phi(x, self._predict_state.v)
        yhat[:, 0] = 1.0 - yhat[:, 1]
        return yhat

    def _reconcile_class(self) -> None:
        '''Make sure models agree about which class is positive.'''
        assert self._my_state is not None, (
            'BUG: self._my_state should not be None when _reconcile_class is called.'
        )
        if all(m.positive_class == self._my_state.positive_class
               for m in self._neighbor_models_iter):
            return  # nothing to do

        old_positive_class = self._my_state.positive_class or 0
        self._my_state.positive_class = int(max(
            [old_positive_class]
            + [m.positive_class for m in self._neighbor_models_iter]
        ))
        # modify models in place
        for m in self._neighbor_models_iter:
            if m.positive_class != self._my_state.positive_class:
                m.v = -m.v
                m.positive_class = self._my_state.positive_class
        if self._my_state.positive_class != old_positive_class and self._my_state is not None:
            self._my_state.v = -self._my_state.v  # pylint: disable=invalid-unary-operand-type

    def _decode(self, serialized_model: bytes) -> AutonLogisticRegressionNeighbor:
        '''Decode a message from distributed neighbors.'''
        return AutonLogisticRegressionNeighbor.decode(serialized_model)

    def _prepare_data(self, dataset: Optional[Dataset]  # pylint: disable=too-many-branches
                      ) -> Tuple[np.ndarray, np.ndarray]:
        '''Extract covariates and target, accounting for the possibility that there's no data.'''
        if dataset is None:
            if list(self._neighbor_models_iter):
                positive_class = int(max(m.positive_class for m in self._neighbor_models_iter))
                d = max(len(m.v) for m in self._neighbor_models_iter) - 1
                np_cov = np.zeros((0, d))
            else:
                # TODO(Piggy/Dan) Need to figure out how to deal with no neighbors and no data.
                raise NoDataError("No data and no neighbors.")
            np_tar = np.zeros(0)
        else:
            target_col_name = dataset.metadata.roles[RoleName.TARGET][0].name
            if DatasetKeys.DATAFRAME.value in dataset:
                orig_cov = dataset[DatasetKeys.DATAFRAME.value].copy()
                if target_col_name in orig_cov.columns:
                    orig_cov.drop(columns=[target_col_name], inplace=True)
                    orig_tar = dataset[DatasetKeys.DATAFRAME.value].loc[:, target_col_name]
                    np_tar = orig_tar.to_numpy()
                else:
                    np_tar = np.array([0])
            else:
                orig_cov = dataset[DatasetKeys.COVARIATES.value]
                if DatasetKeys.TARGET.value in dataset:
                    orig_tar = dataset[DatasetKeys.TARGET.value]
                    np_tar = orig_tar[target_col_name].to_numpy()
                else:
                    np_tar = np.array([0])

            np_cov = orig_cov.to_numpy()

            positive_class = int(max(np_tar))
            if RoleName.TARGET in dataset.metadata.pos_labels:
                try:
                    poslabel = dataset.metadata.pos_labels[RoleName.TARGET]
                    positive_class = int(poslabel)
                except TypeError as e:
                    raise MetadataError(
                        f'{self._algorithm.name if self._algorithm is not None else None} requires '
                        'integer classes, but '
                        'positive label in metadata cannot be converted to int. '
                        f'Found {poslabel} of type {type(poslabel)} instead.'
                    ) from e

        # Add a column for the intercept.
        np_cov = np.hstack((np_cov, np.ones((np_cov.shape[0], 1))))

        _, d = np_cov.shape

        if self._my_state is None:
            self._my_state = AutonLogisticRegressionNeighbor(
                positive_class=positive_class,
                v=np.zeros(d),
                columns=self._columns)

        self._reconcile_class()
        return np_cov, np_tar

    def _objective_with_gradient(
            self, x: np.ndarray, y: np.ndarray, v_old: np.ndarray,
            l2: np.float64, _lambda: np.float64, omega: np.float64,
            k: int) -> Tuple[Callable[[np.ndarray], np.float64],
                             Callable[[np.ndarray], np.ndarray]]:
        """
        Data loss:
            min_v L2_Loss(data,v)
            + L2_coeff*||v||^2
            + lambda*(1-omega)/omega*||v-v0||^2
            + sum_{j in N} lambda/|N|*||v-v_j||
        """

        if k > 0:
            def fun(v: np.ndarray) -> np.float64:
                data_loss = (
                    -(y * _log_phi(x, v)).sum()             # positive class
                    - ((1 - y) * _log_1mphi(x, v)).sum()    # negative class
                )
                l2_regularization = (l2) * norm2(v[:-1])

                # whether to use DJAM or fn space reg
                # if DJAM: need to know weights for all neighbors
                # need a plan to handle 2 cases:
                #   weight from neighbor that we recieved no message from
                #       (set to 0?)
                #   message from neighbor with no weight
                #       (ignore or throw error)

                if self._distributed.regularization_type == 'djam':
                    weights: Dict[NodeID, float] = self._distributed.neighbor_weights
                    distributed_regularization = 0.0
                    for nid, neighbor in self._neighbor_metadata.items():
                        assert isinstance(neighbor.current_state, LinearNeighborState)

                        w = weights[nid]
                        inprod = inner_product(
                            v - neighbor.current_state.v, v - neighbor.current_state.v)
                        distributed_regularization += w * inprod  # type: ignore

                    distributed_regularization = 0.5 * distributed_regularization

                else:
                    neighbor_regularization = (
                        -2.0 * (_lambda / k)
                        * sum(inner_product(m.v, v) for m in self._neighbor_models_iter)
                    )
                    self_regularization = (
                        -2.0 * (_lambda * (1 - omega) / omega) * inner_product(v_old, v)
                    )
                    shared_regularization = (_lambda / omega) * norm2(v)
                    distributed_regularization = (
                        neighbor_regularization + self_regularization + shared_regularization)
                return (data_loss
                        + l2_regularization
                        + distributed_regularization)

            jac = jacobian(fun)  # pylint: disable=no-value-for-parameter

            # def jac(v: np.ndarray) -> np.ndarray:
            #     #This isn't necessarily correct
            #     data_loss = (
            #         -(y[:, np.newaxis] * _grad_log_phi(x, v)).sum(0)
            #         - ((1 - y[:, np.newaxis]) * _grad_log_1mphi(x, v)).sum(0)
            #     )
            #     l2_regularization = (_lambda / omega + l2) * grad_norm2(v)
            #     neighbor_regularization = (
            #         - 2 * _lambda / k
            #         * sum(grad_inner_product(m.v, v) for m in self._neighbor_models_iter)
            #     )
            #     self_regularization = (
            #         - 2 * _lambda * (1 - omega) / omega * grad_inner_product(v_old, v)
            #     )
            #     return (data_loss
            #             + l2_regularization
            #             + neighbor_regularization
            #             + self_regularization)

        else:
            def fun(v: np.ndarray) -> np.float64:
                data_loss = -(y * _log_phi(x, v)).sum() - ((1 - y) * _log_1mphi(x, v)).sum()
                l2_regularization = l2 * norm2(v[:-1])
                return data_loss + l2_regularization

            jac = jacobian(fun)  # pylint: disable=no-value-for-parameter

            # def jac(v: np.ndarray) -> np.ndarray:
            #     print("DEBUG v", v)
            #     data_loss = (
            #         -(y[:, np.newaxis] * _grad_log_phi(x, v)).sum(0)
            #         - ((1 - y[:, np.newaxis]) * _grad_log_1mphi(x, v)).sum(0)
            #     )
            #     l2_regularization = np.array(list(l2 * grad_norm2(v[:-1])) + [0])
            #     return data_loss + l2_regularization

        return fun, jac


class AutonLogisticRegression(Algorithm):
    '''Class for Auton Lab's implementation of Logistic Regression'''
    _name = "auton_logistic_regression"
    _tags = {
        'tasks': [TaskType.BINARY_CLASSIFICATION.name],
        'data_types': [DataType.TABULAR.name],
        'source': ['auton_lab'],
        'distributed': ['true']
    }
    _instance_constructor = AutonLogisticRegressionInstance
    _default_hyperparams = {
        'Lambda': 100.0,
        'L2': 1.0,
        'omega': 2.0 / 3.0,
        'tol': 0.000000001,
        'maxiter': None,
    }

    def instantiate(self, **hyperparams) -> 'AutonLogisticRegressionInstance':
        return super().instantiate(**hyperparams)


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = AutonLogisticRegression(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Available instantiators'''
from typing import Dict, Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..executor.executor_kind import ExecutorKind
from ..executor.simple.simple_instantiator import SimpleInstantiator
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..wrangler.saver import Saver
from .executable_pipeline import ExecutablePipeline
from .instantiator import InstantiatorStub, Instantiator
from .json_instantiator import JsonInstantiator

# TODO(piggy): Maybe replace this with a catalog?
INSTANTIATORS = {
    ExecutorKind('json'): JsonInstantiator,
    ExecutorKind('stub_executor_kind'): InstantiatorStub,
    ExecutorKind('simple'): SimpleInstantiator,
}


class InstantiatorFactory():
    '''Factory class for pipeline instantiators'''
    _kind = None
    _saver: Optional[Saver]

    def __init__(self, saver: Optional[Saver] = None):
        self._saver = saver

    def build(self, kind: ExecutorKind) -> Instantiator:
        '''Build an instantiator of the requested kind.'''
        return INSTANTIATORS[kind](saver=self._saver)

    def instantiate(
        self,
        kind: ExecutorKind,
        pipeline: BoundPipeline
    ) -> ExecutablePipeline:
        '''Instantiate a single bound pipeline into a single kind of executable pipeline.

        Optionally instantiate the pipeline with the given trained models.
        '''
        return INSTANTIATORS[kind](saver=self._saver).instantiate(
            pipeline=pipeline)

    def instantiate_all(
        self,
        kind: ExecutorKind,
        pipelines: Dict[Designator, BoundPipeline]
    ) -> Dict[Designator, ExecutablePipeline]:
        '''Instantiate a dict of bound pipelines into a dict of executable pipelines,

        all of kind 'kind'.'''
        return INSTANTIATORS[kind](saver=self._saver).instantiate_all(
            pipelines=pipelines)



'''Test the PipelineLoaderCatalog'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from .pipeline_loader import PipelineLoaderStub
from .pipeline_loader_catalog import PipelineLoaderCatalogStub

# pylint: disable=missing-function-docstring,duplicate-code


def test_register() -> None:
    dut = PipelineLoaderCatalogStub()
    assert 'stub_pipeline_loader' == dut.register(obj=PipelineLoaderStub())


def test_register_name() -> None:
    dut = PipelineLoaderCatalogStub()
    stub_pipeline_loader = PipelineLoaderStub()
    assert 'somenamedpipelineloader' == dut.register(
        obj=stub_pipeline_loader, name='SomeNamedPipelineLoader')


def test_lookup_by_name() -> None:
    dut = PipelineLoaderCatalogStub()
    some_pipeline_loader = PipelineLoaderStub()
    dut.register(obj=some_pipeline_loader, name='SomeNamedPipelineLoader')
    assert dut.lookup_by_name('SomeNamedPipelineLoader') == some_pipeline_loader


def test_lookup_one_by_tag() -> None:
    dut = PipelineLoaderCatalogStub()
    dut.register(obj=PipelineLoaderStub(), name='NamedStubPipelineLoader', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 1


def test_lookup_by_tag_fail() -> None:
    dut = PipelineLoaderCatalogStub()
    dut.register(obj=PipelineLoaderStub(), name='NamedStubPipelineLoader', tags={'k1': ['v1']})
    assert len(dut.lookup_by_tag_and(k1='different_value')) == 0


def test_lookup_multi_by_tag() -> None:
    dut = PipelineLoaderCatalogStub()
    stub_pipeline_loader = PipelineLoaderStub()
    dut.register(obj=stub_pipeline_loader, name="Name1", tags={'k1': ['v1']})
    stub_pipeline_loader2 = PipelineLoaderStub()
    dut.register(obj=stub_pipeline_loader2, name="Name2", tags={'k1': ['v1']})
    stub_pipeline_loader3 = PipelineLoaderStub()
    dut.register(obj=stub_pipeline_loader3, name="Name3", tags={'k1': ['different_value']})
    assert len(dut.lookup_by_tag_and(k1='v1')) == 2



'''Tests for the Dataset class'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
import pytest

from ..tables.impl.table import TableFactory
from ..tables.impl.table_auto import TableCatalogAuto

from .dataset import Dataset, Metadata, RoleName, Column
from .dataset import DatasetKeys, DatasetKeyError, DatasetValueError

# pylint: disable=missing-function-docstring,pointless-statement,duplicate-code

_ = TableCatalogAuto()


def test_output():
    roles = {RoleName.TIME: Column('bug')}
    dut = Dataset(metadata=Metadata(roles=roles))
    dut['some_key'] = 'some_val'
    got = dut.output().roles
    assert Column('bug') == got[RoleName.TIME]
    with pytest.raises(KeyError):
        got['some_key']


def test_output_overrides():
    roles = {RoleName.TIME: Column('bug')}
    roles2 = {RoleName.TIME: Column('bugs')}
    dut = Dataset(metadata=Metadata(roles=roles))
    dut['some_key'] = 'some_val'
    got = dut.output(override_metadata=Metadata(roles=roles2))
    assert Column('bugs') == got.roles[RoleName.TIME]
    with pytest.raises(KeyError):
        got['some_key']


def test_output_overrides_role():
    roles = {RoleName.TIME: Column('bug')}
    roles2 = {RoleName.TIME: Column('bugs')}
    dut = Dataset(metadata=Metadata(roles=roles))
    dut['some_key'] = 'some_val'
    got = dut.output(override_metadata=dut.metadata.override_roles(roles=roles2))
    assert Column('bugs') == got.roles[RoleName.TIME]


def test_get_dataframe():
    dut = Dataset()
    dataframe = pd.DataFrame({'bug': [1, 2], 'bug2': ['a', 'b']})
    dut.dataframe = dataframe

    assert 'bug' == dut.get_dataframe().columns[0]


def test_df_setters_and_getters() -> None:
    dut = Dataset()
    df1 = pd.DataFrame({'a': [1]})
    df2 = pd.DataFrame({'b': [2]})
    df3 = pd.DataFrame({'c': [3]})
    dut.dataframe = df1
    dut.ground_truth = df2
    dut.predictions = df3
    pd.testing.assert_frame_equal(df1, dut.dataframe)
    pd.testing.assert_frame_equal(df2, dut.ground_truth)
    pd.testing.assert_frame_equal(df3, dut.predictions)


def test_df_getters_proper_errors() -> None:
    dut = Dataset()
    with pytest.raises(DatasetKeyError):
        _ = dut.dataframe

    dut[DatasetKeys.DATAFRAME.value] = 'hamster'
    with pytest.raises(DatasetValueError, match='hamster'):
        _ = dut.dataframe

    dut[DatasetKeys.GROUND_TRUTH.value] = 'gerbil'
    with pytest.raises(DatasetValueError, match='gerbil'):
        _ = dut.ground_truth

    dut[DatasetKeys.PREDICTIONS.value] = 'skink'
    with pytest.raises(DatasetValueError, match='skink'):
        _ = dut.predictions


def test_sorted_columns() -> None:
    '''Test _sort_columns()'''
    input_df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9]})

    input_df = input_df.reindex(columns=['c', 'b', 'a'])
    want_df = input_df.reindex(columns=['a', 'b', 'c'])

    input_covariates_df = input_df[['c', 'b']]
    want_covariates_df = input_covariates_df.reindex(columns=['b', 'c'])

    # Confirm that it matters that the columns are out of order.
    with pytest.raises(AssertionError):
        pd.testing.assert_frame_equal(input_df, want_df)

    dut = Dataset(dataframe=input_df,
                  covariates=input_covariates_df,
                  target=input_df['a'])

    got = dut.sorted_columns()

    pd.testing.assert_frame_equal(got['dataframe'], want_df)
    pd.testing.assert_frame_equal(got['covariates'], want_covariates_df)


def test_table_trio() -> None:
    '''Test dataframe_table, covariates_table, target_table'''
    input_df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9]})
    want_covariates_df = input_df[['b', 'c']]
    want_target_df = input_df['a']

    dut = Dataset(metadata=Metadata(roles={RoleName.TARGET: [Column('a')]}))
    dut.dataframe_table = TableFactory(input_df)

    pd.testing.assert_frame_equal(dut.dataframe_table.as_(pd.DataFrame), input_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long
    pd.testing.assert_frame_equal(dut.covariates_table.as_(pd.DataFrame), want_covariates_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long
    pd.testing.assert_series_equal(dut.target_table.as_(pd.Series), want_target_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long


def test_table_trio_dataframe() -> None:
    '''Test dataframe_table, covariates_table, target_table with dataframe input.'''
    input_df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9]})
    want_covariates_df = input_df[['b', 'c']]
    want_target_df = input_df['a']

    dut = Dataset(metadata=Metadata(roles={RoleName.TARGET: [Column('a')]}))
    dut.dataframe = input_df

    pd.testing.assert_frame_equal(dut.dataframe_table.as_(pd.DataFrame), input_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long
    pd.testing.assert_frame_equal(dut.covariates_table.as_(pd.DataFrame), want_covariates_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long
    pd.testing.assert_series_equal(dut.target_table.as_(pd.Series), want_target_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long


def test_table_trio_dataframe_no_target() -> None:
    '''Test dataframe_table, covariates_table, target_table with dataframe input and no target.'''
    input_df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9]})
    want_covariates_df = input_df[['a', 'b', 'c']]

    dut = Dataset()
    dut.dataframe = input_df

    pd.testing.assert_frame_equal(dut.dataframe_table.as_(pd.DataFrame), input_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long
    pd.testing.assert_frame_equal(dut.covariates_table.as_(pd.DataFrame), want_covariates_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long
    with pytest.raises(DatasetKeyError):
        _ = dut.target_table


def test_trio_dataframe_from_table() -> None:
    '''Test dataframe, covariates_table, target_table with dataframe input from dataframe_table.'''
    input_df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9]})
    want_covariates_df = input_df[['b', 'c']]
    want_target_df = input_df['a']

    dut = Dataset(metadata=Metadata(roles={RoleName.TARGET: [Column('a')]}))
    dut.dataframe_table = TableFactory(input_df)

    pd.testing.assert_frame_equal(dut.dataframe, input_df)
    pd.testing.assert_frame_equal(dut.covariates_table.as_(pd.DataFrame), want_covariates_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long
    pd.testing.assert_series_equal(dut.target_table.as_(pd.Series), want_target_df)  # type: ignore[attr-defined]  # pylint: disable=line-too-long



'''A bound pipeline is a DAG of models with bound hyperparameters'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Any, List, Optional

from ..generator.designator import Designator
from ..searcher.frozen_overrides import FrozenOverrides
from ..templates.impl.parallel_step import ParallelStep
from ..templates.impl.pipeline_step import PipelineStep
from ..templates.impl.pipeline_template import PipelineTemplate


class Error(Exception):
    '''Base error class for BoundPipeline.'''


class StepError(Error):
    '''A step violates BoundPipeline constraints.'''


class ValidationError(Error):
    '''The BoundPipeline object violates constraints.'''


class BoundPipeline(PipelineTemplate):
    '''A ``BoundPipeline`` is a ``PipelineTemplate`` with no query steps.

    ``BoundPipeline``\\ s are outputs of the ``Generator``,
    inputs and outputs of the ``Searcher``,
    and inputs to the ``Instantiator``.
    '''
    _frozen_overrides: Optional[FrozenOverrides] = None

    def __init__(self,
                 name: Optional[str] = None,
                 **kwargs):
        assert name is not None, (
            'BoundPipeline constructor requires a name or designator')
        super().__init__(name=name, **kwargs)
        self.validate()

    @property
    def family_designator(self) -> Designator:
        '''The name this bound pipeline shares when we ignore overrides.'''
        components = [self.name]
        components.extend(
            step.pipeline_designator_component
            for step in self.steps
            if step.queried
        )
        return Designator('@'.join(components).lower())

    @property
    def designator(self) -> Designator:
        '''The unique designator for this bound pipeline'''
        components = [self.name]

        for step in self.steps:
            has_overrides = (self._frozen_overrides is not None
                             and step.pipeline_designator_component in self._frozen_overrides)
            if not has_overrides:
                if not step.queried:
                    # Case: no overrides, not queried.
                    continue
                # Case: no overrides, queried.
                components.append(step.pipeline_designator_component)
                continue
            # Case: has overrides, may or may not be queried.
            assert self._frozen_overrides is not None
            step_overrides = self._frozen_overrides[step.pipeline_designator_component]
            overrides_strs = sorted(
                [f'{k}={v}' for k, v in step_overrides.items()])
            components.append(
                f'{step.pipeline_designator_component}:{",".join(overrides_strs)}')

        return Designator('@'.join(components).lower())

    @property
    def _query_names(self) -> List[str]:
        '''A list of names of all steps that are from a query.

        We currently don't support query steps inside parallel steps.
        '''
        return [step.pipeline_designator_component for step in self._steps if step.queried]

    @classmethod
    def build(cls,
              steps: List[PipelineStep],
              template_name: str,
              frozen_overrides: Optional[FrozenOverrides] = None,
              **kwargs):
        '''Build a BoundPipeline with precalculated steps.'''
        retval = cls(name=template_name, **kwargs)
        retval._steps = steps
        retval._frozen_overrides = frozen_overrides
        return retval

    def new(self, name: str) -> 'BoundPipeline':
        '''Create a new descendant BoundPipeline

        This is for building parallel threads.
        '''
        child = BoundPipeline(name=f'{name}')
        return child

    def validate(self) -> None:
        '''Raise an exception if not everything in the pipeline is bound.'''
        errors: List[Error] = []
        for step in self._steps:
            if not step.has_model():
                if isinstance(step, ParallelStep):
                    for key in step.subpipeline_keys:
                        try:
                            step.subpipeline(key).validate()
                        except Error as err:
                            errors.append(err)
                    continue
                errors.append(StepError(f'step {step.pipeline_designator_component} lacks a model'))
        if errors:
            raise ValidationError(errors)

    def __str__(self):
        return f'{self.designator}'


class BoundPipelineStub(BoundPipeline):
    '''this is a stub'''
    _name = 'stub_pipeline'

    def __init__(self, name: str, tags: Optional[Any] = None, **kwargs):
        tags = tags or {}
        super().__init__(name=name, tags=tags, **kwargs)



'''Tests for sklearn_metrics.py'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Tuple

import numpy as np
import pandas as pd
import pytest

from .impl.metric_catalog import MetricCatalog
from ..metrics.impl.metric import MetricInvalidDatasetError
from ..tables.impl.table import TableFactory
from ..tables.impl.table_auto import TableCatalogAuto
from ..wrangler.dataset import Dataset, Metadata, RoleName, Column

from .sklearn_metrics import register, SklearnMetric
# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code

_ = TableCatalogAuto()  # pylint: disable=pointless-statement


def make_catalog() -> MetricCatalog:
    retval = MetricCatalog()
    register(retval)
    return retval


@pytest.fixture(scope='session')
def classification_results() -> Tuple[Dataset, Dataset]:
    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        pos_labels={RoleName.TARGET: 'a'})
    pred = Dataset(metadata=metadata)
    pred.predictions = pd.DataFrame({'target': ['a', 'b', 'b', 'a']})
    pred.probabilities = TableFactory({'target': [.3, .4, .5, .5]})

    ground_truth = Dataset(metadata=metadata)
    ground_truth.ground_truth = pd.DataFrame({'target': ['a', 'a', 'b', 'a']})
    return (pred, ground_truth)


@pytest.fixture(scope='session')
def regression_results() -> Tuple[Dataset, Dataset]:
    pred = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        )
    )
    pred.predictions = pd.DataFrame({'target': [1.3, 2.8, 7.5, 2.1]})

    ground_truth = Dataset(
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('target')]}
        )
    )
    ground_truth.ground_truth = pd.DataFrame({'target': [0.4, 3.5, 7.6, 2.0]})
    return (pred, ground_truth)


@pytest.mark.parametrize("name,value,precision", [
    ('roc_auc_score', 0.8333333333, 1e-6),
    ('accuracy_score', 0.75, 1e-6),
    ('f1_score', 0.8, 1e-6),
])
def test_classif_metrics(name: str, value: float, precision: float,
                         classification_results: Tuple[Dataset, Dataset]) -> None:
    pred, ground_truth = classification_results
    catalog = make_catalog()
    dut = catalog.lookup_by_name(name=name)

    assert dut.calculate(pred=pred, ground_truth=ground_truth) == pytest.approx(value, precision)


def test_roc_curve(classification_results: Tuple[Dataset, Dataset]) -> None:
    pred, ground_truth = classification_results
    catalog = make_catalog()
    dut = catalog.lookup_by_name(name='roc_curve')
    assert isinstance(dut, SklearnMetric)
    got = dut.calculate_roc_curve(pred=pred, ground_truth=ground_truth)
    want = pd.DataFrame({
        'fpr': [0.0, 1.0 / 3.0, 1.0],
        'tpr': [0.0, 1.0, 1.0],
        'thresholds': [np.inf, 0.5, 0.3]})
    pd.testing.assert_frame_equal(got.dataframe, want)

    with pytest.raises(NotImplementedError):
        dut.calculate(pred=pred, ground_truth=ground_truth)


@pytest.mark.parametrize("name,value,precision", [
    ('mean_absolute_error', 0.45, 1e-6),
    ('mean_squared_error', 0.33, 1e-6),
    ('root_mean_squared_error', 0.57445626, 1e-6),
    ('r2_score', 0.95385825, 1e-6)
])
def test_regression_metrics(name: str, value: float, precision: float,
                            regression_results: Tuple[Dataset, Dataset]) -> None:
    pred, ground_truth = regression_results
    catalog = make_catalog()
    dut = catalog.lookup_by_name(name=name)

    assert dut.calculate(pred=pred, ground_truth=ground_truth) == pytest.approx(value, precision)


def test_roc_auc_no_gt(classification_results: Tuple[Dataset, Dataset]) -> None:
    pred, _ = classification_results
    pred.probabilities = TableFactory({'target': [.3, .4, .5, .5]})
    catalog = make_catalog()
    dut = catalog.lookup_by_name('roc_auc_score')

    with pytest.raises(MetricInvalidDatasetError, match=r'(?i)ground.*truth'):
        dut.calculate(pred=pred)


def test_roc_auc_no_proba(classification_results: Tuple[Dataset, Dataset]) -> None:
    pred, ground_truth = classification_results
    catalog = make_catalog()
    dut = catalog.lookup_by_name('roc_auc_score')
    del pred['probabilities']

    with pytest.raises(MetricInvalidDatasetError, match=r'(?i)probabilit'):
        dut.calculate(pred=pred, ground_truth=ground_truth)


def test_accuracy_no_gt(classification_results: Tuple[Dataset, Dataset]) -> None:
    pred, _ = classification_results
    catalog = make_catalog()
    dut = catalog.lookup_by_name('accuracy_score')

    with pytest.raises(MetricInvalidDatasetError, match=r'(?i)ground.*truth'):
        dut.calculate(pred=pred)



'''Sends and recieves messages using Kafka.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
import logging
from queue import Queue, Full
from typing import Any, Callable, Dict, List, Optional, Set

from aenum import Enum as AEnum  # type: ignore[import-untyped]
from kafka import (  # type: ignore[import-untyped,attr-defined]
    KafkaProducer, KafkaConsumer, KafkaAdminClient, KafkaClient)
from kafka.admin import NewTopic  # type: ignore[import-untyped]


from ..config_components.distributed_config import DistributedConfig, CommunicatorConfig
from ..config_components.impl.config_component import ConfigComponent
from ..neighbor_manager.event import NewNeighbor, Recv
from ..neighbor_manager.node_id import NodeID
from ..wrangler.constants import Defaults
from ..wrangler.exception_thread import ExceptionThread
from ..wrangler.logger import Logger

from .impl.communicator import Communicator
from .impl.communicator_catalog import CommunicatorCatalog

log = Logger(__file__, level=logging.DEBUG).logger()


class KafkaConfig(ConfigComponent):
    '''Specifies Kafka configuration in a distributed AI setting.'''
    name = 'kafka'
    tags: Dict[str, Any] = {}
    _my_id: NodeID

    def __init__(self,
                 clause: Dict[str, Any],
                 name: Optional[str] = None,
                 my_id: NodeID = NodeID(0)) -> None:
        if name is None:
            name = self.name
        super().__init__(name=self.name, clause=clause)
        self._my_id = my_id

    class Constants(Enum):
        '''Keys below the top level.'''
        BOOTSTRAP_SERVERS = 'bootstrap_servers'
        TOPIC = 'topic'
        TOPIC_PREFIX = 'topic_prefix'

    class Keys(AEnum):
        '''Valid keys for the top level.'''
        BOOTSTRAP_SERVERS = 'bootstrap_servers'
        TOPIC = 'topic'

    class Defaults(Enum):
        '''Default values'''
        TOPIC_PREFIX = 'node_'

    def required_keys(self) -> Set[str]:
        return {
            self.Keys.BOOTSTRAP_SERVERS.value,  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            self.Keys.TOPIC.value,  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        }

    @property
    def bootstrap_servers(self) -> List[str]:
        '''List of Kafka servers to connect to.'''
        return self._get(self.Keys.BOOTSTRAP_SERVERS)

    @property
    def topic(self) -> str:
        '''Topic to publish and subscribe to.'''
        return self._get(self.Keys.TOPIC)

    def node_topic(self, node_id: Optional[NodeID] = None) -> str:
        '''Topic for unicast communication.'''
        if node_id is None:
            node_id = self._my_id
        prefix = self._get_with_default(self.Constants.TOPIC_PREFIX,
                                        dflt=self.Defaults.TOPIC_PREFIX.value)
        return f'{prefix}{node_id}'


def _kafka_listener(queue: Queue,
                    stop: Callable[[], bool],
                    timeout: int,
                    known_neighbors: Set[NodeID],
                    consumer: KafkaConsumer,
                    my_id: NodeID) -> None:
    '''Listen for messages on a Kafka topic and put them in a queue.'''
    log.debug('Starting Kafka listener for %s', my_id)
    while True:
        try:
            if stop():
                consumer.close()
                return
            topics = consumer.poll(timeout_ms=timeout)
            consumer.commit()
            for messages in topics.values():
                for message in messages:
                    log.debug('Received message from %s:%s',
                              str(message.key)[:256], str(message.value)[:256])
                    payload = message.value
                    endpoint = int(message.key)
                    neighbor = NodeID(endpoint)
                    # We do not want to read our own messages.
                    if neighbor == my_id:
                        log.debug('Ignoring message from self')
                        continue
                    if neighbor not in known_neighbors:
                        log.debug('Adding new neighbor %s', neighbor)
                        known_neighbors.add(neighbor)
                        queue.put(NewNeighbor(neighbor=neighbor), timeout=timeout)
                    queue.put(Recv(neighbor, payload), timeout=timeout)
        except Full:
            # If the queue fills, drop messages.
            log.debug('Queue is full, dropping message')


class KafkaCommunicator(Communicator):
    '''Sends and recieves messages using Kafka.'''
    name: str = 'kafka'
    tags: Dict[str, List[str]] = {}
    _stop: bool
    _distributed: DistributedConfig
    _admin: Optional[KafkaAdminClient]
    _client: Optional[KafkaClient]
    _consumer: Optional[KafkaConsumer]
    _producer: Optional[KafkaProducer]
    _listener: Optional[ExceptionThread] = None

    def __init__(self,
                 known_neighbors: Set[NodeID],
                 my_id: NodeID,
                 distributed: DistributedConfig):
        self._stop = False
        self._admin = None
        self._client = None
        self._consumer = None
        self._producer = None
        self._distributed = distributed

        super().__init__(my_id=my_id, known_neighbors=known_neighbors, distributed=distributed)

    def stop(self):
        '''Tell threads to terminate themselves.'''
        if self._listener is not None:
            log.debug('Stopping KafkaCommunicator: %s', self._listener)
            self._stop = True
            self._listener.join()
            self._listener = None
        if self._admin is None:
            self._admin.close()
            self._admin = None
        if self._client is not None:
            self._client.close()
            self._client = None
        if self._consumer is not None:
            self._consumer.close()
            self._consumer = None
        if self._producer is not None:
            self._producer.close()
            self._producer = None

    def start(self,
              queue: Queue,
              timeout: float = Defaults.LISTENER_TIMEOUT) -> None:
        '''Start the communicator.'''
        self._client = KafkaClient(
            bootstrap_servers=self._distributed.communicator.kafka.bootstrap_servers  # type: ignore[attr-defined] # pylint: disable=line-too-long
        )
        self._admin = KafkaAdminClient(
            bootstrap_servers=self._distributed.communicator.kafka.bootstrap_servers  # type: ignore[attr-defined] # pylint: disable=line-too-long
        )
        if self._distributed.communicator.strategy == 'broadcast':
            topic = self._distributed.communicator.kafka.topic  # type: ignore[attr-defined]
        else:
            topic = self._distributed.communicator.kafka.node_topic(self._my_id)  # type: ignore[attr-defined] # pylint: disable=line-too-long
        self._consumer = KafkaConsumer(
            topic,
            bootstrap_servers=self._distributed.communicator.kafka.bootstrap_servers,  # type: ignore[attr-defined] # pylint: disable=line-too-long
            client_id=str(self._my_id),
            group_id=str(self._my_id),
            auto_offset_reset='earliest',
            enable_auto_commit=False
        )
        # Create the topics if they do not exist.
        if self._distributed.communicator.strategy == 'broadcast':
            topics = [self._distributed.communicator.kafka.topic]  # type: ignore[attr-defined]
        elif self._distributed.communicator.strategy == 'unicast':
            topics = [self._distributed.communicator.kafka.node_topic(self._my_id)]  # type: ignore[attr-defined] # pylint: disable=line-too-long
            neighbors = self._distributed.get_static_adjacency(self._my_id)
            assert neighbors is not None, 'No neighbors found when creating kafka topics.'
            topics.extend(
                self._distributed.communicator.kafka.node_topic(neighbor)  # type: ignore[attr-defined] # pylint: disable=line-too-long
                for neighbor in neighbors
            )
        for topic in topics:
            log.info('creating topic = %s', topic)
            new_topic = NewTopic(
                name=topic,
                num_partitions=1,
                replication_factor=1)
            future = self._client.cluster.request_update()
            self._client.poll(future=future,
                              timeout_ms=int(1000 * self._distributed.polling_interval))
            metadata = self._client.cluster
            log.debug("metadata.topics() = %s", metadata.topics())
            while topic not in metadata.topics():
                self._admin.create_topics([new_topic])
                future = self._client.cluster.request_update()
                self._client.poll(future=future,
                                  timeout_ms=int(1000 * self._distributed.polling_interval))
                metadata = self._client.cluster
                log.debug("metadata.topics() = %s", metadata.topics())

        self._producer = KafkaProducer(
            bootstrap_servers=self._distributed.communicator.kafka.bootstrap_servers,  # type: ignore[attr-defined] # pylint: disable=line-too-long
            client_id=str(self._my_id),
        )
        self._stop = False

        self._listener = ExceptionThread(target=_kafka_listener, kwargs={
            'queue': queue,
            'stop': lambda: self._stop,
            'timeout': timeout,
            'known_neighbors': self._known_neighbors,
            'consumer': self._consumer,
            'my_id': self._my_id,
        })
        self._listener.start()

    def send(self,
             dest_ids: List[NodeID],
             payload: bytes) -> int:
        '''Send a message to the nodes with the given dest_ids.

        Return value is the total size of the messages sent.
        '''
        assert self._producer is not None, 'Producer is not initialized'
        retval = 0
        # TODO(piggy): If the python API is ever extended to support multiple destinations,
        # we should use it instead of looping.
        for dest_id in dest_ids:
            log.debug('sending to dest_id = %s', dest_id)
            self._producer.send(self._distributed.communicator.kafka.node_topic(dest_id),  # type: ignore[attr-defined] # pylint: disable=line-too-long
                                key=str(self._my_id).encode('utf-8'),
                                value=payload)
            retval += 1
        return retval

    def send_all(self, payload: bytes) -> int:
        '''Send a message to all known neighbors.

        Return value is normally a count of neighbors we sent to,
        but this implementation always returns 1 because we don't know.
        '''
        assert self._producer is not None, 'Producer is not initialized'
        self._producer.send(self._distributed.communicator.kafka.topic,  # type: ignore[attr-defined] # pylint: disable=line-too-long
                            key=str(self._my_id).encode('utf-8'),
                            value=payload)
        return 1


def register(catalog: CommunicatorCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(KafkaCommunicator)
    CommunicatorConfig.register(name='kafka', config_type=KafkaConfig)
